//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the cgi class
//

#pragma implementation "cgi"

#include <cerrno>
#include <cstdarg>
#include <cstdio> // for snprintf
#include <cstdlib>
#include <cstring>
#include <unistd.h>

#include <cgi.h>
#include <configuration.h>
#include <input/crop.h>
#include <input/stdin.h>
#include <mime_lex.h>
#include <rcstring/accumulator.h>

// This is the single Common Gateway Interface onject required for the
// whole program.
cgi gateway;


cgi::~cgi()
{
}


cgi::cgi() :
    variables_set(false),
    slurp_cookie_done(false)
{
    variables.set_reaper();
}


const char *
cgi::getenv(const char *name)
{
    const char *result = ::getenv(name);
    if (!result || !*result)
    {
	fatal_error("Environment variable %s not set", name);
    }
    return result;
}


rcstring
cgi::getenv_or_empty(const char *name)
{
    const char *result = ::getenv(name);
    return (result ? result : "");
}


const rcstring &
cgi::get_script_name()
{
    if (script_name.empty())
    {
	script_name = "http://";
	script_name += getenv("SERVER_NAME");

	const char *SERVER_PORT = getenv("SERVER_PORT");
	int port = atoi(SERVER_PORT);
	if (port > 0 && port < 65536 && port != 80)
	{
	    char buffer[8];
	    snprintf(buffer, sizeof(buffer), ":%d", port);
	    script_name += buffer;
	}
	script_name += getenv("SCRIPT_NAME");
    }
    return script_name;
}


const char *
cgi::get_request_method()
{
    return getenv("REQUEST_METHOD");
}


bool
cgi::viable_path_info()
{
    return !getenv_or_empty("PATH_INFO").empty();
}


const rcstring &
cgi::get_path_info()
{
    if (path_info_cache.empty())
    {
	rcstring result = getenv_or_empty("PATH_INFO");
	if (result.empty())
	    result = "/";
	path_info_cache = result;
	if (path_info_cache.front() != '/')
	    path_info_cache = ("/" + path_info_cache);
    }
    return path_info_cache;
}


void
cgi::fatal_error(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fatal_error_v(fmt, ap);
    va_end(ap);
}


void
cgi::fatal_error_v(const char *fmt, va_list ap)
{
    char buffer[2000];
    vsnprintf(buffer, sizeof(buffer), fmt, ap);

    printf("Content-Type: text/html\n");
    printf("\n");
    printf("<HTML><HEAD><TITLE>Error</TITLE></HEAD><BODY>\n");
    printf("<H1>Error</H1>The following error occurred during\n");
    printf("processing:<BLOCKQUOTE>\n");
    printf("<PRE>%s</PRE>\n", buffer);
    printf("</BLOCKQUOTE></BODY></HTML>\n");
    exit(1);
}


void
cgi::slurp_query_info()
{
    rcstring query(getenv_or_empty("QUERY_STRING"));
    const char *cp = query.c_str();
    while (*cp)
    {
	if (*cp == '&')
	{
	    ++cp;
	    continue;
	}
	const char *amp = strchr(cp, '&');
	if (!amp)
	    amp = cp + strlen(cp);
	const char *eq = (const char *)memchr(cp, '=', amp - cp);
	if (eq)
	{
	    rcstring name(cp, eq - cp);
	    ++eq;
	    rcstring value(eq, amp - eq);
	    set_variable_unquote(name, value);
	}
	else
	{
	    rcstring name(cp, amp - cp);
	    rcstring value("1");
	    set_variable_unquote(name, value);

	    // Turn variables with no value into action=name
	    set_variable_unquote("action", name);
	}
	cp = amp;
    }
    variables_set = true;
}


const rcstring &
cgi::get_variable(const rcstring &name)
{
    if (!variables_set)
	slurp_query_info();

    rcstring *p = variables.query(name);
    if (!p)
    {
	static rcstring nothing;
	return nothing;
    }
    return *p;
}


void
cgi::set_variable(const rcstring &name, const rcstring &value)
{
    variables.assign(name, value);
}


void
cgi::set_variable_unquote(const rcstring &name, const rcstring &value)
{
    variables.assign(name, value.replace("+", " ").url_unquote());
}


void
cgi::read_post_data()
{
    rcstring_accumulator sa;
    rcstring_accumulator name;
    rcstring_accumulator value;
    bool in_value = false;
    input *ip = new input_stdin();
    long limit = config.get_post_limit();
    if (limit > 0)
	ip = new input_crop(ip, limit);
    for (;;)
    {
	int c = ip->getch();
	if (c < 0)
	{
	    if (!name.empty())
		set_variable_unquote(name.mkstr(), value.mkstr());
	    break;
	}

	//
	// Hang onto all of the data for a while yet.
	//
	sa.push_back(c);

	if (c == '&')
	{
	    if (!name.empty())
		set_variable_unquote(name.mkstr(), value.mkstr());
	    in_value = false;
	    name.clear();
	    value.clear();
	}
	else if (c == '=')
	{
	    in_value = true;
	}
	else if (in_value)
	    value.push_back(c);
	else
	    name.push_back(c);
    }
    delete ip;
    set_variable("POST", sa.mkstr());
}


rcstring
cgi::get_remote_host_errok()
{
    return getenv_or_empty("REMOTE_HOST");
}


rcstring
cgi::get_remote_addr()
{
    return getenv("REMOTE_ADDR");
}


rcstring
cgi::get_remote_host_or_ip()
{
    rcstring result = cgi::get_remote_host_errok();
    if (!result.empty())
	return result;
    return get_remote_addr();
}


rcstring
cgi::get_remote_ident_errok()
{
    return getenv_or_empty("REMOTE_IDENT");
}


rcstring
cgi::get_remote_user_errok()
{
    return getenv_or_empty("REMOTE_USER");
}


rcstring
cgi::get_remote_ident_or_user()
{
    rcstring result = get_remote_ident_errok();
    if (!result.empty())
	return result;
    return get_remote_user_errok();
}


rcstring
cgi::get_remote_identification()
{
    rcstring host = get_remote_host_or_ip();
    rcstring user = get_remote_ident_or_user();
    if (user.empty())
	return host;
    return (user + "@" + host);
}


bool
cgi::slurp_cookie_item(mime_lex &line)
{
    if (line.cur_token() != mime_lex::token_name)
	return false;
    rcstring name = line.get_value();
    line.next_token();
    if (line.cur_token() != mime_lex::token_equals)
	return false;
    line.next_token();
    switch (line.cur_token())
    {
    default:
	cookie.assign(name, "");
	break;

    case mime_lex::token_string:
    case mime_lex::token_name:
	cookie.assign(name, line.get_value().c_str());
	line.next_token();
	break;
    }
    switch (line.cur_token())
    {
    case mime_lex::token_end_of_string:
	return true;

    case mime_lex::token_semicolon:
	line.next_token();
	return true;

    default:
	return false;
    }
}


void
cgi::slurp_cookie()
{
    slurp_cookie_done = true;
    mime_lex line(getenv_or_empty("HTTP_COOKIE"));
    line.next_token();
    for (;;)
    {
	switch (line.cur_token())
	{
	case mime_lex::token_end_of_string:
	    return;

	case mime_lex::token_semicolon:
	    line.next_token();
	    break;

	default:
	    if (!slurp_cookie_item(line))
		return;
	    break;
	}
    }
}


rcstring
cgi::get_cookie(const rcstring &name)
{
    if (!slurp_cookie_done)
	slurp_cookie();
    rcstring *p = cookie.query(name);
    return (p ? *p : "");
}
