//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/cgi.cc
//

#ifndef LIB_CGI_H
#define LIB_CGI_H

#pragma interface "cgi"

#include <cstdarg>

#include <format_printf.h>
#include <symtab/template.h>

class mime_lex; // forward

/**
  * The cgi class is used to represent the input state of the current
  * CGI query.
  */
class cgi
{
public:
    /**
      * The destructor.
      */
    virtual ~cgi();

    /**
      * The default constructor.
      */
    cgi();

    /**
      * The get_server_name method is used to obtain the root for
      * all query related URLs.  It contains the protocol, servername
      * (optional port number) and the script name.  The value is cahced
      * for efficient HTML generation.
      */
    const rcstring &get_script_name();

    /**
      * The fatal_error method is used to print a fatal error message on
      * the output and then exit the program.
      */
    void fatal_error(const char *fmt, ...)
							FORMAT_PRINTF(2, 3);
    /**
      * The fatal_error_v method is used to print a fatal error message
      * on the output and then exit the program.  It is called by
      * fatal_error, but it can also be called from outside the class.
      */
    void fatal_error_v(const char *fmt, va_list ap)
							FORMAT_PRINTF(2, 0);
    /**
      * The get_request_method method is used to get the REQUEST_METHOD
      * environment variable.  It is a fatal error if this environment
      * variable is not set.
      */
    const char *get_request_method();

    /**
      * The get_path_info method is used to get the PATH_INFO
      * environment variable.  It is defaulted to "/" if not set.
      */
    const rcstring &get_path_info();

    /**
      * The viable_path_info method is used to determine whether or not
      * we need to issue a Location redirect.
      */
    bool viable_path_info();

    /**
      * The get_variable method is used to determine if the named
      * variable was given in the GET query string, or the POST data if
      * read_post_data method has been called.
      *
      * @param name
      *     The name of the variable to search for.
      * @returns
      *     The data of the named variable, or the empty string if the
      *     variable has not been set.
      */
    const rcstring &get_variable(const rcstring &name);

    /**
      * The set_variable method is used to set the named variable to the
      * given value.  It is used to set the variables found in the query
      * string and the POST data.
      *
      * @param name
      *     The name of the variable to set.
      * @param value
      *     The value of the variable to set.
      */
    void set_variable(const rcstring &name, const rcstring &value);

    /**
      * The read_post_data method is used to read the POST data from
      * the standard input.  The data are populated into the variables
      * instance variable.
      */
    void read_post_data();

    /**
      * The get_remote_host_or_ip method is used to obtain the name, or
      * perhapse only the IP address, of the remote host.
      */
    rcstring get_remote_host_or_ip();

    /**
      * The get_remote_ident_or_user method is used to obtain the name
      * of the remote user.
      */
    rcstring get_remote_ident_or_user();

    /**
      * The get_remote_addr function is used to get the IP address of
      * the remote (client) system.
      */
    rcstring get_remote_addr();

    /**
      * The get_remote_host_errok function is used to get the host name
      * (reverse lookup of IP address) of the remote (client) system.
      * If unavailable, the empty string will be returned.
      */
    rcstring get_remote_host_errok();

    /**
      * The get_remote_ident_errok function is used to get the "ident"
      * value of the remote user.  If unavailable, the empty string will
      * be returned.
      */
    rcstring get_remote_ident_errok();

    /**
      * The get_remote_user_errok function is used to get the user name
      * of the remote user.  If unavailable, the empty string will be
      * returned.
      */
    rcstring get_remote_user_errok();

    /**
      * The get_remote_identification function is used to get a
      * identifying string for the remote user, using the above four
      * methods.  If unavailable, the empty string will be returned.
      */
    rcstring get_remote_identification();

    /**
      * The get_cookie method is used to get the value of the named
      * cookie.
      *
      * @param name
      *     The name of the cookie to look for.
      * @returns
      *     a string containing the cookie value, or the empty string if
      *     the cookie has not been set.
      */
    rcstring get_cookie(const rcstring &name);

private:
    /**
      * The getenv method is used to obtain the value of an environment
      * variable.  It will generate a fatal error if the variable is not
      * set.
      */
    const char *getenv(const char *name);

    /**
      * The getenv_or_empty method is used to obtain the value of an
      * environment variable.  The empty string will be returned if the
      * variable is not set.
      */
    rcstring getenv_or_empty(const char *name);

    /**
      * The server_name instance variable is used to remember the cached
      * value of the root for all query related URLs.  When it isn't
      * empty, It contains the protocol, the servername (optional port
      * number) and the script name.
      */
    rcstring script_name;

    /**
      * The variables instance variable is used to remember the
      * variables in the query string and the POST data.
      */
    symtab<rcstring> variables;

    /**
      * The variables_set instance variable is used to remember whether
      * the slurp_query_info method has been called to populate the
      * variables instance variable.
      */
    bool variables_set;

    /**
      * The slurp_query_info method is used to parse the QUERY_INFO into
      * variable names and values, and populate the variables symbol
      * table.
      */
    void slurp_query_info();

    /**
      * The slurp_cookie method is used by the get_cookie method to read
      * the cookies from the HTTP_COOKIE environment variable into the
      * cookie symbol table.
      */
    void slurp_cookie();

    /**
      * The slurp_cookie_item method is called by the slurp_cookie method
      * to process each cookie definition.
      *
      * @returns
      *     boolean; true if parsed OK, false if there was a parse error.
      */
    bool slurp_cookie_item(mime_lex &);

    /**
      * The slurp_cookie_done instance variable is used to remember
      * whether or not the slurp_cookie method has been called yet.
      */
    bool slurp_cookie_done;

    /**
      * The cookie symbol table is used to remember the values of each
      * cookie as decoded by the slurp_cookie method.
      */
    symtab<rcstring> cookie;

    /**
      * The set_variable_unquote method is used to set the named
      * variable to the given value.  The hexadecimal quoting of
      * characters within the value are decoded first.  It is used to
      * set the variables found in the query string and the POST data.
      *
      * @param name
      *     The name of the variable to set.
      * @param value
      *     The value of the variable to set, once the hexadecimal
      *     escapes have been decoded.
      */
    void set_variable_unquote(const rcstring &name, const rcstring &value);

    /**
      * The path_info_cache is used to remember the results for the
      * get_path_info method.
      */
    rcstring path_info_cache;

    /**
      * The copy constructor.  Do not use.
      */
    cgi(const cgi &);

    /**
      * The assignment operator.  Do not use.
      */
    cgi &operator=(const cgi &);
};

//
// There is only ever one of these objects needed.
//
extern cgi gateway;

#endif // LIB_CGI_H
