//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the configuration class
//

#pragma implementation "configuration"

#include <cassert>
#include <cctype>
#include <cstdio> // for snprintf
#include <cstdlib>
#include <cstring>

#include <boolean.h>
#include <configuration.h>
#include <libdir.h>
#include <output.h>
#include <rcstring/accumulator.h>
#include <rcstring/list.h>


//
// the single instance for the whole program.
//
configuration config;


configuration::~configuration()
{
}


configuration::configuration() :
    sidebar_layout((sidebar_layout_t)0)
{
}


void
configuration::set(const rcstring &name, const rcstring &value)
{
    database.set(name, value);
}


void
configuration::set_from_env(const rcstring &name)
{
    rcstring NAME("QUICKIE_" + name.upcase());
    const char *env = ::getenv(NAME.c_str());
    if (env)
    {
	rcstring value = rcstring(env).trim();
	if (!value.empty())
	    set(name, value);
    }
}


void
configuration::database_read()
{
    if (!database.empty())
	return;

    //
    // Defaults from compile time.
    //
    set("bindir", configured_bindir());
    set("datadir", configured_datadir());
    set("sysconfdir", configured_sysconfdir());
    set("localstatedir", configured_localstatedir());
    set("libdir", configured_libdir());

    //
    // We need to check for an environment variable earlier than all the
    // other wettings.
    //
    set_from_env("sysconfdir");

    //
    // Now read the system configuration file,
    // if it exists.
    //
    rcstring quickie_conf(get_sysconfdir() + "/quickie.conf");
    database.load_from_file(quickie_conf);

    //
    // Defaults from environment variables.
    //
    set_from_env("bindir");
    set_from_env("datadir");
    set_from_env("localstatedir");
    set_from_env("libdir");
}


const rcstring &
configuration::get(const rcstring &name)
{
    //
    // Make sure the database has been read in.
    //
    if (database.empty())
	database_read();

    //
    // Look in the database for the named item.
    //
    return database.get(name);
}


bool
configuration::get_bool(const rcstring &name, bool dflt)
{
    bool result = string_to_bool(get(name), dflt);
    set(name, bool_to_string(result));
    return result;
}


void
configuration::settings(output &os)
{
    //
    // Default a few things, first.
    //
    get_anonymous_user_edit();
    get_chroot();
    get_default_user_edit();
    get_post_limit();
    get_recent_limit();
    get_show_mod_times();
    get_sidebar_layout();
    get_sidebar_width();
    get_style_sheet_name();
    get_umask();

    //
    // Now get all of the values configured (even ones we completely
    // ignore :-) and print them one per line, as a row of an HTML
    // table.
    //
    rcstring_list keys;
    database.keys(keys);
    keys.sort();
    for (size_t j = 0; j < keys.size(); ++j)
    {
	const rcstring &key = keys[j];
	os << "<TR CLASS=\"administration\">"
	      "<TD VALIGN=\"TOP\" CLASS=\"administration\">"
	   << key.capitalize()
	   << ":</TD>"
	      "<TD VALIGN=\"TOP\" CLASS=\"administration\">"
	   << get(key)
	   << "</TD></TR>\n"
	   ;
    }
}


static rcstring
long_to_string(long n, int radix)
{
    char buffer[30];
    if (radix == 16)
	snprintf(buffer, sizeof(buffer), "0x%lX", n);
    else if (radix == 8)
	snprintf(buffer, sizeof(buffer), "0%lo", n);
    else
	snprintf(buffer, sizeof(buffer), "%ld", n);
    return buffer;
}


long
configuration::get_long(const rcstring &name, long dflt)
{
    //
    // Get the string from the database.
    //
    const rcstring &value = get(name);
    if (value.empty())
    {
	oops:
	int radix = (name == "umask" ? 8 : 10);
	set(name, long_to_string(dflt, radix));
	return dflt;
    }

    //
    // Try to turn it into a number.
    //
    char *end = 0;
    long result = strtol(value.c_str(), &end, 0);
    switch (*end++)
    {
    case 'k':
    case 'K':
	result <<= 10;
	break;

    case 'm':
    case 'M':
	result <<= 20;
	break;

    case 'g':
    case 'G':
	result <<= 30;
	break;

    default:
	--end;
	break;
    }
    if (*end)
	goto oops;
    return result;
}


static bool
valid_width(const rcstring &width)
{
    if (width.empty())
	return false;
    const char *s = width.c_str();
    char *ep = 0;
    long n = strtol(s, &ep, 10);
    if (ep == s)
	return false;
    if (0 == strcasecmp(ep, "px"))
    {
	// number of pixels
	return (n > 0 && n < 4000);
    }
    if (ep[0] == '%' && ep[1] == '\0')
    {
	// percentage
	return (n > 0 && n < 100);
    }
    // unknown suffix
    return false;
}


/////// Put all public getters below this line, in alphabetical order //////


bool
configuration::get_anonymous_user_edit()
{
    return get_bool("anonymous-user-edit", false);
}


const rcstring &
configuration::get_bindir()
{
    return get("bindir");
}


const rcstring &
configuration::get_datadir()
{
    return get("datadir");
}


bool
configuration::get_default_user_edit()
{
    return get_bool("default-user-edit", true);
}


const rcstring &
configuration::get_libdir()
{
    return get("libdir");
}


const rcstring &
configuration::get_localstatedir()
{
    if (localstatedir_cache.empty())
	localstatedir_cache = get("localstatedir");
    return localstatedir_cache;
}


long
configuration::get_recent_limit()
{
    return get_long("recent-limit", 100);
}


bool
configuration::get_show_mod_times()
{
    return get_bool("show-mod-times", true);
}


configuration::sidebar_layout_t
configuration::get_sidebar_layout()
{
    if (!sidebar_layout)
    {
	const rcstring value = get("sidebar-layout").downcase();
	if (value == "left")
	    sidebar_layout = sidebar_layout_left;
	else if (value == "right")
	    sidebar_layout = sidebar_layout_right;
	else if (value == "none")
	    sidebar_layout = sidebar_layout_none;
	else
	{
	    // defaults to "left" if we don't understand
	    set("sidebar-layout", "left");
	    sidebar_layout = sidebar_layout_left;
	}
    }
    return sidebar_layout;
}


const rcstring &
configuration::get_style_sheet_name()
{
    if (style_sheet_name.empty())
    {
	style_sheet_name = get("style-sheet-name");
	if (style_sheet_name.empty())
	    style_sheet_name = "default";
	if (!style_sheet_name.ends_with(".css"))
	    style_sheet_name += ".css";
	set("style-sheet-name", style_sheet_name);
    }
    return style_sheet_name;
}


rcstring
configuration::get_sidebar_width()
{
    rcstring width = get("sidebar-width");
    if (!valid_width(width))
    {
	width = "100px";
	set("sidebar-width", width);
    }
    return width;
}


const rcstring &
configuration::get_sysconfdir()
{
    return get("sysconfdir");
}


long
configuration::get_post_limit()
{
    return get_long("post-limit", 100 << 10);
}
