//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/configuration.cc
//

#ifndef LIB_CONFIGURATION_H
#define LIB_CONFIGURATION_H

#pragma interface "configuration"

#include <rfc822.h>

/**
  * The configuration class is used to represent the configuration
  * parameters for the program.
  */
class configuration
{
public:
    /**
      * The destructor.
      */
    virtual ~configuration();

    /**
      * The default constructor.
      */
    configuration();

    /**
      * The get_anonymous_user_edit method is used to determine whether
      * or not anonymous users may edit wiki pages.  Default to false if
      * not set.
      */
    bool get_anonymous_user_edit();

    /**
      * The get_bindir is used to obtain the appropriate directory
      * for executable files.  Defaults to ${prefix}/bin, and may be
      * overriden by the ./configure --bindir option.
      */
    const rcstring &get_bindir();

    /**
      * The get_datadir is used to obtain the appropriate directory for
      * executable files.  Defaults to ${prefix}/share/quickie, and may be
      * overriden by the ./configure --datadir option.
      */
    const rcstring &get_datadir();

    /**
      * The get_default_user_edit method is used to determine whether or
      * not new users may edit wiki pages.  Default to true if not set.
      */
    bool get_default_user_edit();

    /**
      * The get_libdir is used to obtain the appropriate directory for
      * executable files.  Defaults to ${prefix}/lib/quickie, and may be
      * overriden by the ./configure --libdir option.
      */
    const rcstring &get_libdir();

    /**
      * The get_localstatedir is used to obtain the appropriate
      * directory for executable files.  Defaults to ${prefix}/var/quickie,
      * and may be overriden by the ./configure --localstatedir option.
      */
    const rcstring &get_localstatedir();

    /**
      * The get_recent_limit method is used to get the limit on the
      * number of files that will be displayed by the Recent Chanegs
      * page.
      */
    long get_recent_limit();

    /**
      * The get_show_mod_times method is used to get the show-mod-times
      * configuration value.  Defaults to true of not set.
      */
    bool get_show_mod_times();

    enum sidebar_layout_t
    {
	sidebar_layout_left = 1,
	sidebar_layout_right,
	sidebar_layout_none
    };

    /**
      * The get_sidebar_layout method is used to determine the position
      * of the sidebar in the page layout.
      */
    sidebar_layout_t get_sidebar_layout();

    /**
     * The get_sidebar_width method is used to get the width of the
     * sidebar.  Defaults to 100 pixels if not set.
     */
    rcstring get_sidebar_width();

    /**
      * The get_style_sheet_name method is used to get the name of the
      * file in the /admin directory which contains the cascading style
      * sheet for the wiki.  Defaults to "default" if not set.  Will
      * automagically have ".css" added to the end if you omit it.
      */
    const rcstring &get_style_sheet_name();

    /**
      * The get_sysconfdir is used to obtain the appropriate directory
      * for executable files.  Defaults to ${prefix}/etc, and may be
      * overriden by the ./configure --sysconfdir option.
      */
    const rcstring &get_sysconfdir();

    /**
      * The settings method is used by the settings page to print the
      * various settings values.
      */
    void settings(output &os);

    /**
      * The database_read method is used to prepare the database for
      * use.  It reads the relevant environment variables, and reads the
      * quickie.conf file.
      *
      * The databse will only be read once, no matter how many times
      * this method is called.
      *
      * @note
      *     Most clients of this interface will not use this method,
      *     however quickie_prime needs it.
      */
    void database_read();

    /**
      * The set method is used to insert a record into the database.
      *
      * @note
      *     Most clients of this interface will not use this method,
      *     however quickie_prime needs it.
      */
    void set(const rcstring &name, const rcstring &value);

    /**
      * The chroot method is used to determine of the quickie.conf file
      * asks for chroot, and the perform it if so.  Paths to directories
      * are re-written.
      *
      * \note
      *     It's a bit ugly having the chroot inside the configuration
      *     class, but since it requires re-writing of some paths, this
      *     is the smallest foot-print I could come up with.
      */
    void chroot();

    /**
      * The get_post_limit method is used to obtain the maximum amount
      * of data in a POST session.  Defaults to unlimited (0) if not set.
      *
      * @returns
      *     The maximum number of bytes.  Zero means unlimited.
      */
    long get_post_limit();

private:
    /**
      * The database instance variable is used to remeber the name-value
      * pairs for each configuration parameter.
      */
    rfc822 database;

    /**
      * The sidebar_layout instance variable used to the position of the
      * sidebar in the page layout.
      */
    sidebar_layout_t sidebar_layout;

    /**
      * The style_sheet_name instance variable is used to remember the
      * name of the file in the /admin directory which contains the
      * cascading style sheet for the wiki.  Defaults to "default" if
      * not set.  Will automagically have ".css" added to the end if you
      * omit it.
      */
    rcstring style_sheet_name;

    /**
      * The get_bool method is used to get the value of a boolean
      * configuration variable.
      *
      * @param name
      *     The name of the field to get.
      * @param dflt
      *     The default value of the field.
      * @returns
      *     bool; the value of the field.
      */
    bool get_bool(const rcstring &name, bool dflt = false);

    /**
      * The get method is used to get the named configuration parameter.
      *
      * @param name
      *     The name of the configuration item to get.
      * @returns
      *     A reference to the configuration item's value.  You are not
      *     alloed to change it.  If the configuration item does not
      *     exist, the empty string will be returned.
      */
    const rcstring &get(const rcstring &name);

    /**
      * The set_from_env method is used to insert a record into the
      * database if a corresponding environment is set.
      */
    void set_from_env(const rcstring &name);

    /**
      * The get_long method is used to get the value of a configuration
      * field which is an integer.
      *
      * @param name
      *     The name of the field to get.
      * @param dflt
      *     The default value of the field.
      * @returns
      *     long; the value of the field.
      */
    long get_long(const rcstring &name, long dflt);

    /**
      * The chroot method is used to fetch the chroot configuration
      * parameter, which tells you whether or not to chroot before
      * performing any other operation.
      */
    bool get_chroot();

    /**
      * The umask method is used to fetch the umask configuration
      * parameter, which set the file creation mask.  See umask(2) for
      * more information.
      */
    int get_umask();

    /**
      * The localstatedir_cache instance variable is used to cache the
      * database lookup of the localstatedir configuration parameter.
      */
    rcstring localstatedir_cache;

    /**
      * The copy constructor.  Do not use.
      */
    configuration(const configuration &);

    /**
      * The assignment operator.  Do not use.
      */
    configuration &operator=(const configuration &);
};

//
// Only one configuration onject os required fro the whole program,
// and this is it.
//
extern configuration config;

#endif // LIB_CONFIGURATION_H
