//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/file.cc
//

#ifndef LIB_FILE_H
#define LIB_FILE_H

#pragma interface "file"

#include <rcstring.h>

class output; // forward

/**
  * The file class is used to represent an abstreact intercafe to
  * file-like objects which can be queried through the common gateway
  * interface.
  */
class file
{
public:
    /**
      * The destructor.
      */
    virtual ~file();

    /**
      * The default constructor.
      *
      * @param user_path
      *     The user preception of the path of this file.
      */
    file(const rcstring &user_path);

    /**
      * The request method is used to pass on a CGI request.  The
      * REQUEST_METHOD environment variable to determine which request
      * method to call.
      */
    void request();

    /**
      * The factory class method is used to map a pathname to a file
      * class.  Usually it will be a regular file, sometimes it will be
      * a management function of some kind.
      */
    static file *factory(const rcstring &path);

protected:
    /**
      * The get_request method is used to handle REQUEST_METHOD=GET
      * queries to this object.  The default implementation returns a
      * "404 No such file" error.
      */
    virtual void get_request();

    /**
      * The post_request method is used to handle REQUEST_METHOD=POST
      * queries to this object.  The default implementation call the
      * get_request method.
      */
    virtual void post_request();

    /**
      * The head_request method is used to handle REQUEST_METHOD=HEAD
      * queries to this object.  The default implementation returns a
      * "404 No such file" error.
      */
    virtual void head_request();

    /**
      * The layout_header method is used to emit the standard layout
      * header (as controlled by the config file) to the given output
      * stream.
      *
      * @param out
      *     The output stream to be written.
      * @param title
      *     The title of the page.  The title is expected to have
      *     already been turned into well-behaved HTML, including the
      *     necessary <H1>...</H1> markup.
      */
    void layout_header(output &out, const rcstring &title);

    /**
      * The layout footer method is used to emit the standard layout
      * footer (as controlled by the config file) to the given output
      * stream.
      *
      * @param out
      *     The output stream to be written.
      */
    void layout_footer(output &out);

    /**
      * The get_user_path method is used to obtain the user perception
      * of the path of this file.  That is, the path that would be given
      * in the PATH_INFO environment variable.
      */
    const rcstring &get_user_path() const { return user_path; }

private:
    /**
      * The user_path instance variable is used to remember the path of
      * the file, as given in the PATH_INFO environment variable.
      */
    rcstring user_path;

    /**
      * The sidebar method is used to print the contents of the sidebar.
      * It called by layout_header or layout_footer, but not both.
      *
      * @param out
      *     The output stream to be written.
      */
    void sidebar(output &out);

    /**
      * The default constructor.  Do not use.
      */
    file();

    /**
      * The copy constructor.  Do not use.
      */
    file(const file &);

    /**
      * The assignment operator.  Do not use.
      */
    file &operator=(const file &);
};

#endif // LIB_FILE_H
