//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the file_admin class
//

#pragma implementation "file_admin"

#include <cassert>
#include <cstring>
#include <unistd.h>

#include <cgi.h>
#include <configuration.h>
#include <exists_and.h>
#include <file/admin.h>
#include <file/catalogue.h>
#include <file/literal.h>
#include <file/login.h>
#include <file/new_user.h>
#include <file/recent.h>
#include <file/settings.h>
#include <output/stdout.h>


#define SIZEOF(a) (sizeof(a) / sizeof(a[0]))
#define ENDOF(a) ((a) + SIZEOF(a))


file_admin::~file_admin()
{
}


file_admin::file_admin() :
    file("/admin/index")
{
}


static file *
new_file_catalogue()
{
    return new file_catalogue();
}


static file *
new_file_login()
{
    return new file_login();
}


static file *
new_file_new_user()
{
    return new file_new_user();
}


static file *
new_file_recent()
{
    return new file_recent();
}


static file *
new_file_settings()
{
    return new file_settings();
}


struct table_t
{
    const char *name;
    const char *description;
    file *(*func)();
};

static const table_t table[] =
{
    { "catalogue",
	"Obtain a complete listing of all wiki pages in the system.",
	new_file_catalogue },
    { "login",
        "Login to get permissions specific to you, rather than the "
	    "limited anonymous permissions.",
	new_file_login },
    { "new_user",
	"Create a new user.",
	new_file_new_user },
    { "recent",
        "Obtain a summary list of the most recently changed wiki pages "
	    "in the system.",
	new_file_recent },
    { "settings",
        "Obtain a list of configuration settings for the currently "
	    "running wiki engine.",
	new_file_settings },
};


static rcstring
unmangle_images(const rcstring &path)
{
    if (path.ends_with("_css"))
	return path.substring(0, path.size() - 4) + ".css";
    if (path.ends_with("_png"))
	return path.substring(0, path.size() - 4) + ".png";
    return path;
}


file *
file_admin::factory(const rcstring &user_path)
{
    assert(user_path.starts_with("/admin/"));

    //
    // See if the name is in the table.
    //
    const char *np = user_path.c_str() + 7;
    for (const table_t *tp = table; tp < ENDOF(table) ; ++tp)
    {
	if (0 == strcasecmp(np, tp->name))
	{
	    return tp->func();
	}
    }

    //
    // See if the file actually exists.
    // This is where images and CSS files are stashed.
    //
    rcstring file_system_path(
	unmangle_images(config.get_localstatedir() + user_path));
    if (exists_and_is_readable(file_system_path))
	return new file_literal(user_path, file_system_path);

    //
    // When all else fails, show them the index.
    //
    return new file_admin();
}


void
file_admin::get_request()
{
    output_stdout out;
    out << "Content-Type: text/html\n"
           "\n";
    layout_header(out, "<H1 CLASS=\"administration\">Administration</H1>");
    out << "<DL CLASS=\"administration\">\n";
    for (const table_t *tp = table; tp < ENDOF(table) ; ++tp)
    {
	out << "<DT CLASS=\"administration\">\n"
	       "<A HREF=\""
	    << tp->name
	    << "\" CLASS=\"administration\">"
	    << tp->name
	    << "</A></DT>\n"
	       "<DD CLASS=\"administration\">\n"
	    << tp->description
	    << "</DD>\n";
    }
    out << "</DL>\n";
    layout_footer(out);
}
