//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the file_regular class
//

#pragma implementation "file_regular"

#include <cassert>
#include <cerrno>
#include <cstdlib>
#include <cstdio> // for rename
#include <ctime>
#include <fcntl.h>
#include <unistd.h>

#include <careful.h>
#include <cgi.h>
#include <configuration.h>
#include <file/regular.h>
#include <input/crlf.h>
#include <input/file.h>
#include <input/string.h>
#include <input/svt_checkout.h>
#include <link_tracker/null.h>
#include <link_tracker/wiki.h>
#include <output/file.h>
#include <output/html_quote.h>
#include <output/stdout.h>
#include <quit.h>
#include <relative_path.h>
#include <rfc822_functor/html_history.h>
#include <simple_version_tool.h>
#include <user.h>
#include <wiki_to/html.h>
#include <wiki_word.h>


file_regular::~file_regular()
{
}


file_regular::file_regular(const rcstring &arg1, const rcstring &arg2) :
    file(arg1),
    file_system_path(arg2)
{
    assert(get_user_path()[0] == '/');
    assert(file_system_path[0] == '/');
}


void
file_regular::synthesize()
{
    //
    // Open the files.
    //
    // FIXME: how do we lock out simultaneous updates?
    //
    rcstring html_file_name = file_system_path + ".html";
    rcstring html_file_name_new = html_file_name + "-new";
    output_file op(html_file_name_new);

    time_t when = 0;
    rcstring wiki_file_name = file_system_path + ".wiki";
    input_file ip(wiki_file_name);
    ip.empty_if_absent();
    when = ip.get_mod_time();
    link_tracker_wiki linker(get_user_path());
    wiki_to_html translator("content", ip, op, linker);

    //
    // Extract the first heading from the wiki file; or if none is
    // present, use the user path, but slightly prettied up.
    //
    rcstring heading = translator.look_for_heading_one();
    if (heading.empty())
    {
	heading =
	    (
		"<H1 CLASS=\"content\">"
	    +
		wiki_word_pretty(get_user_path()).html_quote()
	    +
		"</H1>"
	    );
    }

    //
    // Add the standard header.
    //
    layout_header(op, heading);

    //
    // Translate the text.
    //
    translator.translate();

    //
    // At the foot of the page we add some navigation buttons to allow
    // editing the page.  Eventually we should also have history links,
    // and other administrative stuff.
    //
    op.printf("<HR CLASS=\"content\">\n");
    // Note: do NOT url_quote the script name, you mangle the colon
    op << "<P CLASS=\"content\"><A HREF=\""
       << get_user_path().basename()
       << "?edit\" CLASS=\"content\">Edit this page</A>.\n"
          "<A HREF=\""
       << get_user_path().basename()
       << "?history\" CLASS=\"content\">Historical versions</A>.";
    if (config.get_show_mod_times())
    {
	op << "<BR>\n";
	if (when)
	{
	    op << "This content last modified ";
	    op.write(ctime(&when), 24);
	    op << ".\n";
	}
	op << "This html text generated ";
	time(&when);
	op.write(ctime(&when), 24);
	op << '.';
    }
    op << "</P>\n";

    //
    // Add the standard footer.
    //
    layout_footer(op);

    //
    // Now do an atomic replace, so the user never sees half an update.
    //
    careful_rename(html_file_name_new, html_file_name);
}


void
file_regular::get_request_edit()
{
    // If this is empty, use whatever is there at the moment.
    rcstring version = gateway.get_variable("version");

    output_stdout op;
    op << "Content-Type: text/html\n\n";
    rcstring title =
       	(
	    "<H1 CLASS=\"administration\">Edit "
	+
	    wiki_word_pretty(get_user_path()).html_quote()
	+
	    "</H1>"
	);
    layout_header(op, title);

    if (!user.may_edit())
    {
	if (user.is_anonymous())
	{
	    op << "<P CLASS=\"warning\">"
		  "You must <A HRED=\""
	       << relative_path(get_user_path(), "/admin/login")
	       << "\">login</A> to edit wiki pages."
		  "</P>\n";
	}
	else
	{
	    op << "<P CLASS=\"warning\">"
		  "You do not have permission to edit this wiki page."
		  "</P>\n";
	}
	layout_footer(op);
	return;
    }

    op << "<FORM ACTION=\""
       << gateway.get_script_name().c_str()
       << get_user_path().html_quote()
       << "\" METHOD=\"POST\">\n"
          "<P CLASS=\"administration\">\n"
          "<TEXTAREA NAME=\"content\" ROWS=\"24\" COLS=\"80\">";

    //
    // The output needs to escape the HTML special characters, so that
    // the browser will present them to the user correctly.
    //
    output_html_quote out2(op);

    if (!version.empty())
    {
	//
	// If a version was specified, extract the version from history.
	//
	rcstring history_file_path = file_system_path + ",svt";
	input_svt_checkout src(history_file_path, version);
	out2 << src;
    }
    else
    {
	//
	// If no version was specified, use the .wiki file.
	//
	rcstring wiki_file_path = file_system_path + ".wiki";
	input_file src(wiki_file_path);
	src.empty_if_absent();
	out2 << src;
    }

    //
    // Flush this one before we write any more output,
    // or it will come out scrambled.
    //
    out2.flush();

    op << "</TEXTAREA><BR>\n"
          "<INPUT TYPE=\"SUBMIT\" name=\"action\" value=\"Save\">\n"
          "<INPUT TYPE=\"SUBMIT\" name=\"action\" value=\"Preview\">\n"
          "<INPUT TYPE=\"SUBMIT\" name=\"action\" value=\"Cancel\">\n"
          "</P>\n"
          "</FORM>\n";

    //
    // We also print some wiki text format hints.
    //
    rcstring hints_wiki_path =
	config.get_localstatedir() +  "/manual/markup_hints.wiki";
    input_file hints(hints_wiki_path);
    hints.empty_if_absent();
    link_tracker_null no_links(get_user_path());
    wiki_to_html translator("administration", hints, op, no_links);
    translator.translate();

    //
    // Print the standard footer.
    //
    layout_footer(op);
}


void
file_regular::get_request_history()
{
    output_stdout op;
    op << "Content-Type: text/html\n\n";
    rcstring title =
	(
	    "<H1 CLASS=\"administration\">History of "
	+
	    wiki_word_pretty(get_user_path()).html_quote()
	+
	    "</H1>"
	);
    layout_header(op, title);

    if (!user.may_edit())
    {
	op << "<P CLASS=\"administration\">"
	      "You may not edit wiki pages.";
	if (user.is_anonymous())
	{
	    op << "\n"
		  "You may need to <a HREF=\""
	       << relative_path(get_user_path(), "/admin/login")
	       << "\">login</A>\n"
	          "or create a <A HREF=\""
	       << relative_path(get_user_path(), "/admin/new_user")
	       << "\">new account</A>.\n";
	}
	op << "</P>\n";
    }

    //
    // Print the table of previous versions.
    //
    op << "<TABLE CLASS=\"administration\">\n";
    op << "<TR CLASS=\"administration\">\n"
	  "<TH CLASS=\"administration\">Version</TH>\n"
	  "<TH CLASS=\"administration\">Date and Time</TH>\n"
	  "<TH CLASS=\"administration\">Who</TH>\n"
	  "<TH CLASS=\"administration\">Revert?</TH>\n"
	  "</TR>\n";
    simple_version_tool archive(file_system_path + ",svt");
    rfc822_functor_html_history doodle(get_user_path(), op);
    archive.list(doodle);
    op << "</TABLE>\n";

    //
    // Print the standard footer.
    //
    layout_footer(op);
}


static void
warning(output &out)
{
    out << "<P CLASS=\"warning\">Warning: this is a preview.  Your\n"
           "editing has not been saved.  To save, hit the Back button\n"
           "in your broswer, and then click on the Save button.</P>\n";
}


void
file_regular::get_request_preview()
{
    output_stdout out;
    out << "Content-Type: text/html\n"
           "\n";
    rcstring title =
       	(
	    "<H1 CLASS=\"administration\">Preview "
	+
	    wiki_word_pretty(get_user_path()).html_quote()
	+
	    "</H1>"
	);
    layout_header(out, title);

    warning(out);
    out << "<HR>\n";

    //
    // Translate the content.
    //
    const rcstring &content = gateway.get_variable("content");
    input_string ip(content);
    input_crlf ipnl(ip);
    link_tracker_null noref(get_user_path());
    wiki_to_html translator("preview", ipnl, out, noref);
    translator.translate();

    out << "<HR>\n";
    warning(out);
    layout_footer(out);
}


void
file_regular::get_request_html()
{
    rcstring html_file_name = file_system_path + ".html";
    int fd = open(html_file_name.c_str(), O_RDONLY);
    if (fd < 0)
    {
	synthesize();
	fd = open(html_file_name.c_str(), O_RDONLY);
	if (fd < 0)
	{
	    quitter.fatal_error_with_errno
    	    (
		"open \"%s\"",
		html_file_name.c_str()
	    );
	    // NOTREACHED
	}
    }
    struct stat st;
    if (fstat(fd, &st) < 0)
    {
	quitter.fatal_error_with_errno("stat \"%s", html_file_name.c_str());
	// NOTREACHED
    }
    if (!S_ISREG(st.st_mode))
    {
	quitter.fatal_error("%s: Not a regular file", html_file_name.c_str());
	return;
    }
    output_stdout op;
    op << "Content-Type: text/html\n"
          "Content-Length: "
       << st.st_size
       << "\n"
          "\n";
    for (;;)
    {
	char buffer[1 << 14];
	int n = read(fd, buffer, sizeof(buffer));
	if (n <= 0)
	    break;
	op.write(buffer, n);
    }
}


void
file_regular::get_request()
{
    rcstring action = gateway.get_variable("action").downcase();
    if (action == "edit")
	get_request_edit();
    else if (action == "history")
	get_request_history();
    else if (action == "save")
	get_request_save();
    else if (action == "preview")
	get_request_preview();
    else if (action == "cancel")
	get_request_cancel();
    else
	get_request_html();
}


void
file_regular::get_request_save()
{
    if (!user.may_edit())
    {
	output_stdout op;
	op << "Content-Type: text/html\n\n";
	rcstring title =
	    (
		"<H1 CLASS=\"administration\">Save "
	    +
		wiki_word_pretty(get_user_path()).html_quote()
	    +
		"</H1>"
	    );
	layout_header(op, title);
	op << "<P CLASS=\"warning\">"
	      "You do not have permission to save this wiki page."
	      "</P>\n";
	    layout_footer(op);
	return;
    }

    // We need the braces for scope.
    // The .wiki file must be closed before the HTML is regenerated.
    rcstring wiki_file_name = file_system_path + ".wiki";
    rcstring wiki_file_name_new = wiki_file_name + "-new";
    {
	output_file dest(wiki_file_name_new);
	dest.mkdir_p();
	input_string content(gateway.get_variable("content"));
	input_crlf content_minus_crlf(content);
	dest << content_minus_crlf;
	if (!dest.start_of_line())
	    dest << '\n';
    }

    //
    // Now do an atomic replace, so the user never sees half an update.
    //
    careful_rename(wiki_file_name_new, wiki_file_name);

    //
    // Build some meta-data to accompany the checkin.  We remember as
    // much as this so that we can (some time in the future) blacklist
    // the author, if we need to.
    //
    rfc822 meta_data;
    meta_data.set_minimalist("remote-addr", gateway.get_remote_addr());
    meta_data.set_minimalist("remote-host", gateway.get_remote_host_errok());
    meta_data.set_minimalist("remote-ident", gateway.get_remote_ident_errok());
    meta_data.set_minimalist("remote-user", gateway.get_remote_user_errok());
    if (user.is_anonymous())
	meta_data.set_minimalist("user", gateway.get_remote_identification());
    else
	meta_data.set_minimalist("user", user.get_name());

    //
    // Commit the latest verion to the per-file archive.
    //
    simple_version_tool archive(file_system_path + ",svt");
    archive.checkin(wiki_file_name, meta_data);

    //
    // Unlink the HTML file to force it to be regenerated.
    //
    rcstring html_file_name = file_system_path + ".html";
    careful_unlink(html_file_name);

    //
    // Now treat it like a normal request.
    // This causes the HTML to be regenerated and cached.
    //
    get_request_html();
}


void
file_regular::get_request_cancel()
{
    get_request_html();
}
