//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the input class
//

#pragma implementation "input"

#include <cassert>
#include <cerrno>
#include <cstdio> // for vsnprintf
#include <cstring>

#include <input.h>
#include <quit.h>


input::~input()
{
    delete [] buffer;
    buffer = 0;
    buffer_position = 0;
    buffer_high_water = 0;
    buffer_size = 0;
}


input::input() :
    buffer(0),
    buffer_position(0),
    buffer_high_water(0),
    buffer_size(0)
{
}


int
input::underflow()
{
    if (!buffer)
    {
	buffer_size = get_block_size();
	buffer = new unsigned char [buffer_size];
	buffer_position = 0;
	buffer_high_water = 0;
    }
    if (buffer_position >= buffer_high_water)
    {
	buffer_position = 0;
	buffer_high_water = read_inner(buffer, buffer_size);
	if (!buffer_high_water)
	    return -1;
    }
    return buffer[buffer_position++];
}


size_t
input::get_block_size()
    const
{
    return 1024;
}


void
input::ungetc(unsigned char c)
{
    if (buffer_high_water == 0)
    {
	assert(buffer != 0);
	assert(buffer_position == 0);
	assert(buffer_size != 0);
	buffer_high_water = buffer_size;
	buffer_position = buffer_high_water;
	--buffer_position;
	buffer[--buffer_position] = c;
	return;
    }

    assert(buffer_position > 0);
    assert(buffer_position <= buffer_high_water);
    if (buffer_position > 0)
    {
	--buffer_position;
	buffer[buffer_position] = c;
    }
}


bool
input::gets(char *buf, size_t buflen)
{
    if (buflen < 2)
	return false;
    char *begin = buf;
    for (;;)
    {
	int c = getch();
	if (c < 0)
	{
	    if (buf == begin)
		return false;
	    *buf = '\0';
	    return true;
	}
	*buf++ = c;
	--buflen;
	if (c == '\n')
	{
	    *buf = '\0';
	    return true;
	}
    }
}


size_t
input::read(void *void_data_p, size_t data_size)
{
    char *data_start = (char *)void_data_p;
    char *data = data_start;
    char *data_end = data + data_size;
    size_t block_size = get_block_size();
    while (data < data_end)
    {
	if (buffer_position < buffer_high_water)
	{
	    // use buffer contents
	    size_t nbytes = buffer_high_water - buffer_position;
	    if (nbytes > (size_t)(data_end - data))
		nbytes = data_end - data;
	    memcpy(data, buffer + buffer_position, nbytes);
	    data += nbytes;
	    buffer_position += nbytes;
	    if (buffer_position >= buffer_high_water)
	    {
		buffer_position = 0;
		buffer_high_water = 0;
	    }
	}
	else if ((size_t)(data_end - data) >= block_size)
	{
	    // skip the middle man
	    size_t nbytes = read_inner(data, block_size);
	    if (nbytes == 0)
		break;
	    data += nbytes;
	}
	else
	{
	    // underflow
	    if (!buffer)
	    {
		buffer_size = get_block_size();
		buffer = new unsigned char [buffer_size];
		buffer_position = 0;
		buffer_high_water = 0;
	    }
	    buffer_high_water = read_inner(buffer, buffer_size);
	    if (!buffer_high_water)
		break;
	    // loop again, the first case will be used
	}
    }
    return (data - data_start);
}


void
input::fatal_error(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fatal_error_v(fmt, ap);
    va_end(ap);
}


void
input::fatal_error_v(const char *fmt, va_list ap)
{
    char buf[1000];
    vsnprintf(buf, sizeof(buf), fmt, ap);
    quitter.fatal_error("%s: %s", name().c_str(), buf);
}


void
input::fatal_error_with_errno(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fatal_error_with_errno_v(fmt, ap);
    va_end(ap);
}


void
input::fatal_error_with_errno_v(const char *fmt, va_list ap)
{
    int err = errno;
    char buf[1000];
    vsnprintf(buf, sizeof(buf), fmt, ap);
    errno = err;
    quitter.fatal_error_with_errno("%s: %s", name().c_str(), buf);
}


bool
input::at_end()
{
    int c = getch();
    if (c < 0)
	return true;
    ungetc((unsigned char)c);
    return false;
}
