//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/input.cc
//

#ifndef LIB_INPUT_H
#define LIB_INPUT_H

#pragma interface "input"

#include <cstdarg>
#include <cstddef>

#include <format_printf.h>
#include <rcstring.h>

/**
  * The input class is used to represent and abstract input source.  It
  * could be a file, or a string, or something else entirely.
  */
class input
{
public:
    /**
      * The destructor.
      */
    virtual ~input();

    /**
      * The default constructor.
      */
    input();

    /**
      * The getch method is used to get the next character from the
      * input stream.
      *
      * @returns
      *     The next character from the input stream (in the unsigned
      *     char range), or -1 at the end of input.
      */
    int
    getch()
    {
	if (buffer_position < buffer_high_water)
	    return buffer[buffer_position++];
	return underflow();
    }

    /**
      * The ungetc method is used to return a characters recently
      * returned by the getch method.
      *
      * @note
      *     It has to have been read <i>very</i> recently.  You can only
      *     put back the character actually read, an only <i>one</i> is
      *     guaranteed to work.
      */
    void ungetc(unsigned char c);

    /**
      * The ungetc method is used to return a characters recently
      * returned by the getch method.
      */
    void ungetc(signed char c) { ungetc((unsigned char)c); }

    /**
      * The ungetc method is used to return a characters recently
      * returned by the getch method.
      */
    void ungetc(char c) { ungetc((unsigned char)c); }

    /**
      * The ungetc method is used to return a characters recently
      * returned by the getch method.
      */
    void ungetc(int c) { ungetc((unsigned char)c); }

    /**
      * The gets method is used to read a line from the input into the
      * buffer until either a terminating newline or EOF, or the buffer
      * is full.  If the end of file is reached, false is returned.  The
      * buffer is always terminated with a NUL when true is returned.
      *
      * @param buffer
      *     Where to place the data.
      * @param buflen
      *     The maximum number of bytes to place into the supplied buffer.
      */
    bool gets(char *buffer, size_t buflen);

    /**
      * The read method may be used to read blocks if data from the input.
      *
      * @param buffer
      *     Where to place the data.
      * @param buflen
      *     The maximum number of bytes to place into the supplied buffer.
      */
    size_t read(void *buffer, size_t buflen);

    /**
      * The name method is used to obtain the name of the input.
      */
    virtual rcstring name() = 0;

    /**
      * The fatal_error method is used to report fatal errors.
      * The `fmt' string is in the same style a standard C printf
      * function.  It calls the fatal_error_v method.  This method
      * does not return.
      */
    virtual void fatal_error(const char *fmt, ...)
							FORMAT_PRINTF(2, 3);
    /**
      * The fatal_error_v method is used to report fatal errors.
      * The `fmt' string is in the same style a standard C vprintf
      * function.  It calls the message_v and exit methods.
      * This method does not return.
      */
    virtual void fatal_error_v(const char *, va_list)
						        FORMAT_PRINTF(2, 0);
    /**
      * The at_en method is used to determine whether the current input
      * position is at the end of input.
      */
    bool at_end();

protected:
    /**
      * The read_inner method is used to read data in the buffer.
      *
      * @param data
      *     Where to place the data read from the input stream.
      * @param nbytes
      *     The maximum number of bytes to read in to the area pointed
      *     at by the \a data parameter.
      * @returns
      *     The number of characters read into the buffer, or zero at
      *     the end of input.
      * @note
      *     If there is a read error it will be reported via
      *     quit::fatal_error, and the method will not return.
      */
    virtual size_t read_inner(void *data, size_t nbytes) = 0;

    /**
      * The get_block_size method is used to determine the size of the
      * data buffer for caching the data.
      */
    virtual size_t get_block_size() const;

    /**
      * The fatal_error_errno method is used to report fatal errors,
      * and append the string equivalent of errno.	The `fmt' string
      * is in the same style a standard C printf function.	It calls
      * the fatal_error_errno_v method.  This method does not return.
      */
    virtual void fatal_error_with_errno(const char *fmt, ...)
							FORMAT_PRINTF(2, 3);
    /**
      * The fatal_error_errno_v method is used to report fatal
      * errors.  The `fmt' string is in the same style a standard C
      * vprintf function.  It calls the message_v and exit methods.
      * This method does not return.
      */
    virtual void fatal_error_with_errno_v(const char *, va_list)
						        FORMAT_PRINTF(2, 0);
private:
    /**
      * The buffer instance variable is used to remember the address
      * of the base of the dynamically allocated bufefr to hold input
      * chartacters was they are dispensed one at a time by the getch
      * method.
      */
    unsigned char *buffer;

    /**
      * The buffer_position instance variable is used to remember where
      * withing the buffer array the next character is to be read from.
      */
    size_t buffer_position;

    /**
      * The buffer_high_water instance variable is used to remember
      * how many characters have been read into the buffer.  This is
      * different to (less than or equal to) the buffer_size.
      */
    size_t buffer_high_water;

    /**
      * The buffer_size instance variable is used to remember the number
      * of characters allocated to the buffer array.
      */
    size_t buffer_size;

    /**
      * The underflow method is used by the getch method when the data
      * buffer is empty.
      *
      * @returns
      *     The next character from the input stream (in the unsigned
      *     char range), or -1 at the end of input.
      */
    int underflow();

    /**
      * The copy constructor.  Do not use.
      */
    input(const input &);

    /**
      * The assignment operator.  Do not use.
      */
    input &operator=(const input &);
};

#endif // LIB_INPUT_H
