//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/link_tracker.cc
//

#ifndef LIB_LINK_TRACKER_H
#define LIB_LINK_TRACKER_H

#pragma interface "link_tracker"

#include <rcstring.h>

/**
  * The link_tracker class is used to represent an abstract interface to
  * an object that is used to track links between documents.
  */
class link_tracker
{
public:
    /**
      * The destructor.
      */
    virtual ~link_tracker();

    /**
      * The constructor.
      *
      * @param user_path
      *     The path of the file, as given in the PATH_INFO environment
      *     variable.
      */
    link_tracker(const rcstring &user_path);

    /**
      * The internal_link_exists method is sued to notify the tracker
      * of an internal link between two wiki pages, where the \a
      * destination exists.
      *
      * @param destination
      *     A user-perception of the path to the destination wiki page.
      */
    virtual void internal_link_exists(const rcstring &destination) = 0;

    /**
      * The internal_link_new method is sued to notify the tracker
      * of an internal link between two wiki pages, where the \a
      * destination does not yet exist.
      *
      * @param destination
      *     A user-perception of the path to the destination wiki page.
      */
    virtual void internal_link_new(const rcstring &destination) = 0;

    /**
      * The external_link method is sued to notify the tracker of a link
      * between an internal wiki page and an external object.  It is not
      * known whether the \a destination exists or not.
      *
      * @param destination
      *     A user-perception of the path to the destination wiki page.
      */
    virtual void external_link(const rcstring &destination);

    /**
      * The get_user_path method is used to obtain the user perception
      * of the path of this file.  That is, the path that would be given
      * in the PATH_INFO environment variable.
      */
    const rcstring &get_user_path() const { return user_path; }

private:
    /**
      * The user_path instance variable is used to remember the path of
      * the file, as given in the PATH_INFO environment variable.
      */
    rcstring user_path;

    /**
      * The default constructor.  Do not use.
      */
    link_tracker();

    /**
      * The copy constructor.  Do not use.
      */
    link_tracker(const link_tracker &);

    /**
      * The assignment operator.  Do not use.
      */
    link_tracker &operator=(const link_tracker &);
};

#endif // LIB_LINK_TRACKER_H
