//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the link_tracker_wiki class
//

#pragma implementation "link_tracker_wiki"

#include <cerrno>
#include <unistd.h>

#include <careful.h>
#include <configuration.h>
#include <link_tracker/wiki.h>


link_tracker_wiki::~link_tracker_wiki()
{
    //
    // First we need to read in the list of other wiki pages we *used*
    // to reference.  It is what changed that is of most interest.
    //
    rcstring file_system_path = config.get_localstatedir() + get_user_path();
    rcstring_list xref_old_was;
    rcstring refers_to_path = file_system_path + ".refers_to";
    xref_old_was.read_from(refers_to_path);

    //
    // Write out the current list of wiki pages we reference.
    //
    xref_old.double_buffered_write_to(refers_to_path);

    //
    // For each of the pages we used to reference, but no longer
    // reference; OR each page we now reference but did not previously;
    // remove their .html cache file, because its "referred to by" list
    // (in the side bar) is now out of date.
    //
    rcstring_list interesting = xref_old ^ xref_old_was;
    for (size_t j = 0; j < interesting.size(); ++j)
    {
	const rcstring &other_user_path = interesting[j];
	rcstring other_file_system_path =
	    config.get_localstatedir() + other_user_path;
	rcstring other_html = other_file_system_path + ".html";
	careful_unlink(other_html);
    }

    //
    // If this page references any pages which don't exist yet, then
    // it's cached .html file will need to be removed once they do.  So
    // for each of these files, we update it's .unlink_on_create file
    // by appending out user_path.
    //
    for (size_t k = 0; k < xref_new.size(); ++k)
    {
	const rcstring &other_user_path = xref_new[k];
	rcstring other_file_system_path =
	    config.get_localstatedir() + other_user_path;
	rcstring unlink_path = other_file_system_path + ".unlink_on_create";
	rcstring_list back_ref;
	back_ref.read_from(unlink_path);
	back_ref.push_back_unique(get_user_path());
	back_ref.double_buffered_write_to(unlink_path);
    }

    //
    // If this user_path has an .unlink_on_create file, we need to read
    // it, act on it, and delete it.
    //
    rcstring unlink_on_create = file_system_path + ".unlink_on_create";
    rcstring_list nukem;
    nukem.read_from(unlink_on_create);
    for (size_t m = 0; m < nukem.size(); ++m)
    {
	const rcstring &other_user_path(nukem[m]);
	rcstring other_file_system_path =
	    config.get_localstatedir() + other_user_path;
	rcstring html_path = other_file_system_path + ".html";
	careful_unlink(html_path);
    }
    careful_unlink(unlink_on_create);

    //
    // update the referred_to_by files
    //
    interesting = xref_old - xref_old_was;
    for (size_t n = 0; n < interesting.size(); ++n)
    {
	const rcstring &other_user_path = interesting[n];
	rcstring other_file_system_path =
	    config.get_localstatedir() + other_user_path;
	rcstring other_path = other_file_system_path + ".referred_to_by";
	rcstring_list back_ref;
	back_ref.read_from(other_path);
	back_ref.push_back_unique(get_user_path());
	back_ref.double_buffered_write_to(other_path);
    }

    //
    // Need to extract ourself from lists we aren't on any more
    //
    interesting = xref_old_was - xref_old;
    for (size_t p = 0; p < interesting.size(); ++p)
    {
	const rcstring &other_user_path = interesting[p];
	rcstring other_file_system_path =
	    config.get_localstatedir() + other_user_path;
	rcstring other_path = other_file_system_path + ".referred_to_by";
	rcstring_list back_ref;
	back_ref.read_from(other_path);
	back_ref.remove(get_user_path());
	back_ref.double_buffered_write_to(other_path);
    }
}


link_tracker_wiki::link_tracker_wiki(const rcstring &arg) :
    link_tracker(arg)
{
}


void
link_tracker_wiki::internal_link_exists(const rcstring &destination)
{
    xref_old.push_back_unique(destination);
}


void
link_tracker_wiki::internal_link_new(const rcstring &destination)
{
    xref_new.push_back_unique(destination);
    xref_old.push_back_unique(destination);
}
