//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the output class
//

#pragma implementation "output"

#include <cerrno>
#include <cstdarg>
#include <cstdio>
#include <cstring>

#include <output.h>
#include <input.h>
#include <quit.h>
#include <rcstring.h>


output::~output()
{
}


output::output() :
    buffer(0),
    position(0),
    maximum(0)
{
}


void
output::puts(const char *s)
{
    if (s)
	write(s, strlen(s));
}


void
output::puts(const rcstring &s)
{
    write(s.c_str(), s.size());
}


size_t
output::get_block_size()
{
    return 1024;
}


void
output::overflow(unsigned char c)
{
    if (!buffer)
    {
	maximum = get_block_size();
	buffer = new unsigned char [maximum];
    }
    if (position >= maximum)
	flush();
    buffer[position++] = c;
}


void
output::write(const void *data, size_t nbytes)
{
    if (!nbytes)
	return;
    char *p = (char *)data;
    if (!buffer)
    {
	maximum = get_block_size();
	buffer = new unsigned char [maximum];
    }
    for (;;)
    {
	if (position >= maximum)
	{
	    write_inner(buffer, position);
	    position = 0;
	}
	if (!nbytes)
	    break;
	if (position == 0 && nbytes >= maximum)
	{
	    // Don't double copy if we can avoid it.
	    write_inner(p, maximum);
	    p += maximum;
	    nbytes -= maximum;
	}
	else
	{
	    size_t max = maximum - position;
	    if (max > nbytes)
		max = nbytes;
	    memcpy(buffer + position, p, max);
	    position += max;
	    p += max;
	    nbytes -= max;
	}
    }
}


void
output::flush()
{
    if (position)
    {
	write_inner(buffer, position);
	position = 0;
    }
}


void
output::printf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    char temp[2000];
    vsnprintf(temp, sizeof(temp), fmt, ap);
    va_end(ap);
    puts(temp);
}


bool
output::start_of_line()
    const
{
    if (!position)
	return start_of_line_inner();
    return (buffer[position - 1] == '\n');
}


void
output::put(input &from)
{
    for (;;)
    {
	char temp[1024];
	size_t nbytes = from.read(temp, sizeof(temp));
	if (!nbytes)
	    break;
	write(temp, nbytes);
    }
}


void
output::fatal_error(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fatal_error_v(fmt, ap);
    va_end(ap);
}


void
output::fatal_error_v(const char *fmt, va_list ap)
{
    char buf[1000];
    vsnprintf(buf, sizeof(buf), fmt, ap);
    quitter.fatal_error("%s: %s", name().c_str(), buf);
}


void
output::fatal_error_with_errno(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fatal_error_with_errno_v(fmt, ap);
    va_end(ap);
}


void
output::fatal_error_with_errno_v(const char *fmt, va_list ap)
{
    int err = errno;
    char buf[1000];
    vsnprintf(buf, sizeof(buf), fmt, ap);
    errno = err;
    quitter.fatal_error_with_errno("%s: %s", name().c_str(), buf);
}


void
output::put_long_le(unsigned long x)
{
    putch((unsigned char)x);
    putch((unsigned char)(x >> 8));
    putch((unsigned char)(x >> 16));
    putch((unsigned char)(x >> 24));
}
