//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/output.cc
//

#ifndef LIB_OUTPUT_H
#define LIB_OUTPUT_H

#pragma interface "output"

#include <cstdarg>
#include <cstddef>

#include <format_printf.h>
#include <rcstring.h>

class input; // forward

/**
  * The output class is used to represent an abstract output
  * destination.  It could be a file, or a string or something more
  * interesting.
  */
class output
{
public:
    /**
      * The destructor.
      */
    virtual ~output();

    /**
      * The default constructor.
      */
    output();

    /**
      * The putch method is used to write a character to the output.
      *
      * @param c
      *     The character to write.
      */
    void putch(char c) { putch((unsigned char)c); }

    /**
      * The putch method is ised to write a character to the output.
      *
      * @param c
      *     The character to write.
      */
    void putch(signed char c) { putch((unsigned char)c); }

    /**
      * The putch method is ised to write a character to the output.
      *
      * @param c
      *     The character to write.
      * @note
      *     All of these output class buffer their output.  It lowers
      *     the total number of function calls per character quite
      *     dramatically.
      */
    void
    putch(unsigned char c)
    {
	if (position < maximum)
	    buffer[position++] = c;
	else
	    overflow(c);
    }

    /**
      * The puts method is used to write a NUL-terminated string to the
      * output.
      */
    void puts(const char *s);

    /**
      * The puts method is used to write a reference counted string to
      * the output.
      */
    void puts(const rcstring &s);

    /**
      * The write method is used to write arbitrary data to the output.
      *
      * @param data
      *     The data to write.
      * @param nbytes
      *     The home much data to write.
      */
    void write(const void *data, size_t nbytes);

    /**
      * The flush method is used to ensure that all the buffered data
      * has been written to the actual output.
      */
    void flush();

    /**
      * The printf method is used to print formatted text to the
      * output strteam.  See printf(3) for more information about the
      * arguments.
      *
      * @param fmt
      *     The format string which describes the following arguments,
      *     and how to format them.
      */
    void printf(const char *fmt, ...)                       FORMAT_PRINTF(2, 3);

    /**
      * The start_of_line method may be used to determine whether the
      * output is currently positioned at the start of a line.  This can
      * be useful when trying to determine of a Unix text file needs to
      * have its terminating newline appended.
      *
      * @returns
      *     boolean; true if currently positionad at the start of a
      *     line, false if currentl;y position within an output line.
      */
    virtual bool start_of_line() const;

    /**
      * The put method is used to copy an input to an output.
      *
      * @param from
      *     Where the input comes from.
      */
    void put(input &from);

    /**
      * The name method may be used to obtain the name of the output
      * file being written.
      */
    virtual rcstring name() = 0;

    /**
      * The put_long_le is used to write a four byte long value to the
      * output stream in little endian format (least significant byte
      * first).
      *
      * @param x
      *     The value to be written to the output.
      */
    void put_long_le(unsigned long x);

protected:
    /**
      * The write inner method is used to write the buffered data out
      * to the underlying implemnattion.  This method must be supplied
      * by each derived class.
      *
      * @param data
      *     The data to write.
      * @param nbytes
      *     The home much data to write.
      */
    virtual void write_inner(const void *data, size_t nbytes) = 0;

    /**
      * The get_buffer_size method is used to obtain the preferred
      * buffer size of output.  This is not called from the base class
      * destructor, but is deferred until the first data is written to
      * the output.
      *
      * This defaults to 1024, which isn't too bad for sockets, but is
      * not ideal for files.
      */
    virtual size_t get_block_size();

    /**
      * The start_of_line_inner method is called by the start_of_line
      * method in the rare cases where it cannot determine the output
      * state for itself.
      *
      * @returns
      *     boolean; true if currently positionad at the start of a
      *     line, false if currentl;y position within an output line.
      */
    virtual bool start_of_line_inner() const = 0;

    /**
      * The fatal_error method is used to report fatal errors.
      * The `fmt' string is in the same style a standard C printf
      * function.  It calls the fatal_error_v method.  This method
      * does not return.
      *
      * @param fmt
      *     The format string which describes the following arguments,
      *     and how to format them.
      */
    virtual void fatal_error(const char *fmt, ...)
							FORMAT_PRINTF(2, 3);
    /**
      * The fatal_error_v method is used to report fatal errors.
      * The `fmt' string is in the same style a standard C vprintf
      * function.  It calls the message_v and exit methods.
      * This method does not return.
      *
      * @param fmt
      *     The format string which describes the following arguments,
      *     and how to format them.
      * @param args
      *     Variable argument lsit, controlled by the \a fmt argument.
      */
    virtual void fatal_error_v(const char *fmt, va_list args)
						        FORMAT_PRINTF(2, 0);
    /**
      * The fatal_error_errno method is used to report fatal errors,
      * and append the string equivalent of errno.  The `fmt' string
      * is in the same style a standard C printf function.  It calls
      * the fatal_error_errno_v method.  This method does not return.
      *
      * @param fmt
      *     The format string which describes the following arguments,
      *     and how to format them.
      */
    virtual void fatal_error_with_errno(const char *fmt, ...)
							FORMAT_PRINTF(2, 3);
    /**
      * The fatal_error_errno_v method is used to report fatal
      * errors.  The `fmt' string is in the same style a standard C
      * vprintf function.  It calls the message_v and exit methods.
      * This method does not return.
      *
      * @param fmt
      *     The format string which describes the following arguments,
      *     and how to format them.
      * @param args
      *     Variable argument lsit, controlled by the \a fmt argument.
      */
    virtual void fatal_error_with_errno_v(const char *fmt, va_list args)
						        FORMAT_PRINTF(2, 0);
private:
    /**
      * The buffer instance variable is used to remember the address of
      * a dynamically allocated array of characters to buffer the output.
      */
    unsigned char *buffer;

    /**
      * The position instance variable is used to remember the current
      * output position (high water) within the buffer.
      */
    size_t position;

    /**
      * The maximum instance variable is used to remember how many bytes
      * have been allocated in the buffer.
      */
    size_t maximum;

    /**
      * The overflow method is called by the putch method when there is
      * not enough space left in the buffer to store another character.
      *
      * @param c
      *     The character to be written to the output.
      */
    void overflow(unsigned char c);

    /**
      * The copy constructor.  Do not use.
      */
    output(const output &);

    /**
      * The assignment operator.  Do not use.
      */
    output &operator=(const output &);
};

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, const rcstring &s)
{
    op.puts(s);
    return op;
}

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, const char *s)
{
    op.puts(s);
    return op;
}

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, char c)
{
    op.putch(c);
    return op;
}

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, signed char c)
{
    op.putch(c);
    return op;
}

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, unsigned char c)
{
    op.putch(c);
    return op;
}

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, int n)
{
    op.printf("%d", n);
    return op;
}

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, unsigned n)
{
    op.printf("%u", n);
    return op;
}

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, long n)
{
    op.printf("%ld", n);
    return op;
}

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, unsigned long n)
{
    op.printf("%lu", n);
    return op;
}

/**
  * The << operator is used to send data to the output in a type-safe manner.
  */
inline output &
operator<<(output &op, input &rhs)
{
    op.put(rhs);
    return op;
}

#endif // LIB_OUTPUT_H
