//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the output_gzip class
//

#pragma implementation "output_gzip"

#include <ac/zlib.h>

#include <output/gzip.h>

#ifndef Z_BUFSIZE
#ifdef MAXSEG_64K
#define Z_BUFSIZE 4096 // minimize memory usage for 16-bit DOS
#else
#define Z_BUFSIZE 16384
#endif
#endif

static unsigned char gzip_magic[2] = {0x1f, 0x8b}; // gzip magic header


output_gzip::~output_gzip()
{
    //
    // Make sure all buffered data has been passed to our write_inner
    // method.
    //
    flush();

    //
    // finish sending the compressed stream
    //
    stream.avail_in = 0; // should be zero already anyway
    if (stream.avail_out == 0)
    {
	    deeper->write(outbuf, Z_BUFSIZE);
	    stream.next_out = outbuf;
	    stream.avail_out = Z_BUFSIZE;
    }
    for (;;)
    {
	int err = deflate(&stream, Z_FINISH);
	if (err < 0)
	    drop_dead(err);
	uInt len = Z_BUFSIZE - stream.avail_out;
	if (!len)
	    break;
	deeper->write(outbuf, len);
	stream.next_out = outbuf;
	stream.avail_out = Z_BUFSIZE;
    }

    //
    // and the trailer
    //
    deeper->put_long_le(crc);
    deeper->put_long_le(stream.total_in);

    //
    // Clean up any resources we were using.
    //
    if (stream.state != NULL)
	deflateEnd(&stream);
    delete [] outbuf;
    outbuf = 0;

    //
    // Finish the deeper stream.
    //
    if (close_on_close)
	delete deeper;
    deeper = 0;
}


output_gzip::output_gzip(output *arg1, bool arg2) :
    deeper(arg1),
    close_on_close(arg2),
    outbuf(new Byte [Z_BUFSIZE]),
    crc(0),
    pos(0),
    bol(true)
{
    crc = crc32(0L, Z_NULL, 0);
    stream.avail_in = 0;
    stream.avail_out = 0;
    stream.next_in = NULL;
    stream.next_out = NULL;
    stream.opaque = (voidpf)0;
    stream.zalloc = (alloc_func)0;
    stream.zfree = (free_func)0;

    //
    // Set the parameters for the compression.
    // Note: windowBits is passed < 0 to suppress zlib header.
    //
    int err =
	deflateInit2
	(
	    &stream,
	    Z_BEST_COMPRESSION, // level
	    Z_DEFLATED,	        // method
	    -MAX_WBITS,		// windowBits
	    DEF_MEM_LEVEL,      // memLevel
	    Z_DEFAULT_STRATEGY  // strategy
	);
    if (err != Z_OK)
	drop_dead(err);

    stream.next_out = outbuf;
    stream.avail_out = Z_BUFSIZE;

    //
    // Write a very simple .gz header:
    //
    deeper->putch(gzip_magic[0]);
    deeper->putch(gzip_magic[1]);
    deeper->putch((char)Z_DEFLATED);
    deeper->putch((char)0); // flags
    deeper->put_long_le(0L); // time
    deeper->putch((char)0); // xflags
    deeper->putch((char)3); // always use unix OS_CODE
}


void
output_gzip::drop_dead(int err)
{
    fatal_error("gzip: %s", z_error(err));
}


rcstring
output_gzip::name()
{
    return deeper->name();
}


void
output_gzip::write_inner(const void *buf, size_t len)
{
    if (len > 0)
	bol = (((const char *)buf)[len - 1] == '\n');
    stream.next_in = (Bytef *)buf;
    stream.avail_in = len;
    while (stream.avail_in != 0)
    {
	if (stream.avail_out == 0)
	{
	    deeper->write(outbuf, Z_BUFSIZE);
	    stream.next_out = outbuf;
	    stream.avail_out = Z_BUFSIZE;
	}
	int err = deflate(&stream, Z_NO_FLUSH);
	if (err != Z_OK)
	    drop_dead(err);
    }
    crc = crc32(crc, (Bytef *)buf, len);
    pos += len;
}


bool
output_gzip::start_of_line_inner()
    const
{
    return bol;
}
