//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005, 2006 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the post_install_speil function.
//

#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <fcntl.h>
#include <unistd.h>

#include <libdir.h>
#include <path_join.h>
#include <post_install.h>
#include <quit.h>
#include <rcstring/list.h>

static rcstring user;
static rcstring group;
static rcstring cgi_bin;


static void
grope(const rcstring &path)
{
    FILE *fp = fopen(path.c_str(), "r");
    if (!fp)
	return;
    for (;;)
    {
	char buffer[1000];
	if (!fgets(buffer, sizeof(buffer), fp))
	    break;
	rcstring_list arg;
	arg.split(buffer, 0, true);
	if (user.empty() && arg[0] == "User")
	    user = arg[1];
	if (group.empty() && arg[0] == "Group")
	    user = arg[1];
	if (cgi_bin.empty() && arg[0] == "ScriptAlias" && arg[1] == "/cgi-bin/")
	{
	    cgi_bin = arg[2];
	    if (cgi_bin[0] == '"')
		cgi_bin = cgi_bin.substring(1, cgi_bin.size() - 2);
	}
    }
}


static void
rummage(const rcstring &path)
{
    if (!user.empty() && !group.empty() && !cgi_bin.empty())
	return;
    DIR *dp = opendir(path.c_str());
    if (!dp)
	return;
    rcstring_list names;
    for (;;)
    {
	struct dirent *dep = readdir(dp);
	if (!dep)
	    break;
	const char *np = dep->d_name;
	if (np[0] == '.' && (!np[1] || (np[1] == '.' && !np[2])))
	    continue;
	names.push_back(rcstring(np));
    }
    closedir(dp);

    for (size_t j = 0; j < names.size(); ++j)
    {
	const rcstring &name(names[j]);
	rcstring path2(path_join(path, name));
	struct stat st;
	if (lstat(path2.c_str(), &st) < 0)
	    return;
	switch (st.st_mode & S_IFMT)
	{
	case S_IFREG:
	    grope(path2);
	    break;

	case S_IFDIR:
	    rummage(path2);
	    break;
	}
	if (!user.empty() && !group.empty() && !cgi_bin.empty())
	    break;
    }
}


static void
rummage_all()
{
    rummage("/etc/apache2");
    rummage("/etc/apache");
    rummage("/etc/httpd");
}


static bool
compare_data_equal(const rcstring &fn1, const rcstring &fn2)
{
    int fd1 = open(fn1.c_str(), O_RDONLY);
    if (fd1 < 0)
	quitter.fatal_error_with_errno("open %s", fn1.c_str());
    int fd2 = open(fn2.c_str(), O_RDONLY);
    if (fd2 < 0)
	quitter.fatal_error_with_errno("open %s", fn2.c_str());
    for (;;)
    {
	char buf1[4096];
	int n1 = read(fd1, buf1, sizeof(buf1));
	if (n1 < 0)
	    quitter.fatal_error_with_errno("read %s", fn1.c_str());
	if (n1 <= 0)
	    break;
	char buf2[sizeof(buf1)];
	int n2 = read(fd2, buf2, sizeof(buf2));
	if (n2 < 0)
	    quitter.fatal_error_with_errno("read %s", fn2.c_str());
	if
       	(
	    n1 != n2
	||
	    0 != memcmp(buf1, buf2, n1)
	)
	{
	    close(fd1);
	    close(fd2);
	    return false;
	}
    }
    close(fd2);
    close(fd1);
    return true;
}


enum cmp_t
{
    cmp_ok,
    cmp_rm_cp,
    cmp_cp,
    cmp_rm_ln_s,
    cmp_ln_s
};

static cmp_t
files_compare(const rcstring &from, const rcstring &to)
{
    //
    // Get the information on the destination file.
    // It is OK if it does not exist yet.
    //
    struct stat to_st;
#ifdef S_ISLNK
    if (lstat(to.c_str(), &to_st) < 0)
    {
	if (errno != ENOENT)
	    quitter.fatal_error_with_errno("stat %s", to.c_str());
	return cmp_ln_s;
    }
    if (S_ISLNK(to_st.st_mode))
    {
	//
	// If it is a symbolic link, make sure it points to the right place.
	//
	char buffer[2000];
	int n = readlink(to.c_str(), buffer, sizeof(buffer));
	if (n < 0)
	    quitter.fatal_error_with_errno("readlink %s", to.c_str());
	if (rcstring(buffer, n) != from)
	    return cmp_rm_ln_s;
	return cmp_ok;
    }
#else
    if (stat(to.c_str(), &to_st) < 0)
    {
	if (errno != ENOENT)
	    quitter.fatal_error_with_errno("stat %s", to.c_str());
	return cmp_cp;
    }
#endif

    //
    // Get the information about the origin file.
    // It is supposed to exist already.
    //
    struct stat from_st;
    if (stat(from.c_str(), &from_st) < 0)
	quitter.fatal_error_with_errno("stat %s", from.c_str());

    //
    // Both files exist.  Make sure they are identical.
    //
    if (from_st.st_dev == to_st.st_dev && from_st.st_ino == to_st.st_ino)
	return cmp_ok;
    if (from_st.st_size != to_st.st_size || !compare_data_equal(from, to))
    {
#ifdef S_ISLNK
	return cmp_rm_ln_s;
#else
	return cmp_rm_cp;
#endif
    }

    //
    // Nothing needs to be done.
    //
    return cmp_ok;
}


static bool
is_a_number(const rcstring &s, int &n)
{
    char *end = 0;
    n = strtoul(s.c_str(), &end, 10);
    return (end > s.c_str() && *end == 0);
}


static bool
ownership_ok(const rcstring &dir)
{
    //
    // Get information about the wiki data directory.
    //
    struct stat st;
    if (stat(dir.c_str(), &st) < 0)
	quitter.fatal_error_with_errno("stat %s", dir.c_str());
    if (!S_ISDIR(st.st_mode))
    {
	errno = ENOTDIR;
	quitter.fatal_error_with_errno("stat %s", dir.c_str());
    }

    //
    // Turn symbolic user name into numeric UID.
    //
    int uid = st.st_uid;
    if (!is_a_number(user, uid))
    {
	struct passwd *pwd = getpwnam(user.c_str());
	if (!pwd)
	    quitter.fatal_error("user \"%s\" unknown", user.c_str());
	uid = pwd->pw_uid;
    }

    //
    // Turn symbolic group name into numeric GID.
    //
    int gid = st.st_gid;
    if (!group.empty() && !is_a_number(group, gid))
    {
	struct group *grp = getgrnam(group.c_str());
	if (!grp)
	    quitter.fatal_error("group \"%s\" unknown", group.c_str());
	gid = grp->gr_gid;
    }

    //
    // See if onwership is OK.
    //
    return ((int)st.st_uid == uid && (int)st.st_gid == gid);
}


void
post_install_speil()
{
    //
    // Rummage through some likely places looking for Apache
    // configuration data.
    //
    rummage_all();

    if (cgi_bin.empty())
    {
	printf
	(
	    "\n"
	    "You need to copy the quickie executable into your web server's\n"
	    "cgi-bin directory.  This program is unable to work out where\n"
	    "that is.\n"
	);
    }
    else
    {
	rcstring from = path_join(configured_bindir(), "quickie");
	rcstring to = path_join(cgi_bin, "quickie");
	switch (files_compare(from, to))
	{
	case cmp_ok:
	    break;

	case cmp_rm_ln_s:
	    printf
	    (
		"\n"
		"You need to replace the symbolic link to the quickie\n"
		"executable in your web server's cgi-bin directory.\n"
		"\n"
	    );
	    printf("\trm -f %s\n", to.c_str());
	    printf("\tln -s %s %s\n", from.c_str(), to.c_str());
	    goto explain_why;

	case cmp_ln_s:
	    printf
	    (
		"\n"
		"You need to create a symbolic link to the quickie executable\n"
		"into your web server's cgi-bin directory.  (By default,\n"
		"Apache is configured to allow symbolic links in the cgi-bin\n"
		"directory.)\n"
		"\n"
	    );
	    printf("\tln -s %s %s\n", from.c_str(), to.c_str());
	    goto explain_why;

	case cmp_rm_cp:
	    printf
	    (
		"\n"
		"You need to replace the old quickie executable into\n"
		"your web server's cgi-bin directory.  This isn't done\n"
		"automagically by 'make install' because it tends to move\n"
		"around a great deal from one web server to another.\n"
		"\n"
	    );
	    printf("\trm -f %s\n", to.c_str());
	    printf("\tcp %s %s\n", from.c_str(), to.c_str());
	    goto explain_why;

	case cmp_cp:
	    printf
	    (
		"\n"
		"You need to copy the quickie executable into your web\n"
		"server's cgi-bin directory.\n"
		"\n"
	    );
	    printf("\tcp %s %s\n", from.c_str(), to.c_str());
	explain_why:
	    printf
	    (
		"\n"
		"This isn't done automagically by 'make install' because the\n"
		"cgi-bin directory tends to move around a great deal from\n"
		"one web server to another, and one site to another.\n"
		"\n"
		"If you have multiple virtual web servers configured, these\n"
		"commands may be incorrect or insufficient.\n"
	    );
	    break;
	}
    }

    if (user.empty())
    {
	printf
	(
	    "\n"
	    "You need to change the ownership of the quickie wiki data\n"
	    "directories to match the UID your web server executes as.  This\n"
	    "program is unable to work out what UID that is.\n"
	);
    }
    else
    {
	rcstring dir = configured_localstatedir();
	if (!ownership_ok(dir))
	{
	    printf
	    (
		"\n"
		"You need to change the ownership of the quickie data\n"
		"directories to match the UID your web server executes as.\n"
		"\n"
	    );
	    rcstring as(group.empty() ? user : (user + ":" + group));
	    printf("\tchown -R %s %s\n", as.c_str(), dir.c_str());
	    printf
	    (
		"\n"
		"If you have multiple virtual servers configured, this\n"
		"command may be incorrect or insufficient.\n"
	    );
	}
    }
}


static void
execute(const rcstring &cmd)
{
    int n = system(("sh -x -c " + cmd.quote_shell()).c_str());
    if (n != 0)
    {
	quitter.fatal_error
       	(
	    "command failed (exit status %d)",
	    (n | (n >> 8)) & 255
	);
    }
}


void
post_install_cgi_bin()
{
    rummage_all();

    if (cgi_bin.empty())
	quitter.fatal_error("unable to locate web server's cgi-bin directory");

    rcstring from = path_join(configured_bindir(), "quickie");
    rcstring to = path_join(cgi_bin, "quickie");
    switch (files_compare(from, to))
    {
    case cmp_ok:
	quitter.message("%s already correct", to.c_str());
	break;

    case cmp_rm_ln_s:
	execute
	(
	    "rm -f " + to.quote_shell() + ";"
	    "ln -s " + from.quote_shell() + " " + to.quote_shell()
	);
	break;

    case cmp_ln_s:
	execute("ln -s " + from.quote_shell() + " " + to.quote_shell());
	break;

    case cmp_rm_cp:
	execute
       	(
	    "rm -f " + to.quote_shell() + ";"
	    "cp " + from.quote_shell() + " " + to.quote_shell()
	);
	break;

    case cmp_cp:
	execute("cp " + from.quote_shell() + " " + to.quote_shell());
	break;
    }
}


void
post_install_localstatedir()
{
    rummage_all();

    if (user.empty())
	quitter.fatal_error("unable to determine web server UID");
    rcstring dir = configured_localstatedir();
    if (ownership_ok(dir))
    {
	quitter.message("%s ownership already correct", dir.c_str());
    }
    else
    {
	rcstring as(group.empty() ? user : (user + ":" + group));
	execute("chown -R " + as.quote_shell() + " " + dir.quote_shell());
    }
}
