//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/quit.cc
//

#ifndef LIB_QUIT_H
#define LIB_QUIT_H

#pragma interface "quit"

#include <cstdarg>
#include <format_printf.h>
#include <quit/atexit_list.h>

/**
  * The quit class is an abstract class for reporting error messages,
  * both fatal and non-fatal.
  */
class quit
{
public:
    /**
      * The destructor.
      */
    virtual ~quit();

    /**
      * The default constructor.
      */
    quit();

    /**
      * The copy constructor.
      */
    quit(const quit &);

    /**
      * The assignment operator.
      */
    quit &operator=(const quit &);

    /**
      * The fatal_error method is used to report fatal errors.
      * The `fmt' string is in the same style a standard C printf
      * function.  It calls the fatal_error_v method.  This method
      * does not return.
      */
    virtual void fatal_error(const char *fmt, ...)
					       FORMAT_PRINTF(2, 3) NORETURN;
    /**
      * The fatal_error_v method is used to report fatal errors.
      * The `fmt' string is in the same style a standard C vprintf
      * function.  It calls the message_v and exit methods.
      * This method does not return.
      */
    virtual void fatal_error_v(const char *fmt, va_list ap)
					       FORMAT_PRINTF(2, 0) NORETURN;
    /**
      * The fatal_error_errno method is used to report fatal errors,
      * and append the string equivalent of errno.	The `fmt' string
      * is in the same style a standard C printf function.	It calls
      * the fatal_error_errno_v method.  This method does not return.
      */
    virtual void fatal_error_with_errno(const char *fmt, ...)
					       NORETURN FORMAT_PRINTF(2, 3);
    /**
      * The fatal_error_errno_v method is used to report fatal
      * errors.  The `fmt' string is in the same style a standard C
      * vprintf function.  It calls the message_v and exit methods.
      * This method does not return.
      */
    virtual void fatal_error_with_errno_v(const char *fmt, va_list ap)
					       FORMAT_PRINTF(2, 0) NORETURN;
    /**
      * The warning method is used to likely but non-fatal errors.
      * The `fmt' string is in the same style a standard C printf
      * function.  It calls the warning_v method.
      */
    virtual void warning(const char *fmt, ...)
						        FORMAT_PRINTF(2, 3);
    /**
      * The warning_v method is used to report likely but non-fatal
      * errors.  The `fmt' string is in the same style a standard
      * C vprintf function.  It calls the message_v method.
      */
    virtual void warning_v(const char *fmt, va_list ap)
						        FORMAT_PRINTF(2, 0);
    /**
      * The message method is used to send an error message to a
      * suitable destination.  The `fmt' string is in the same style
      * a standard C printf function.  The message_v method is called.
      */
    virtual void message(const char *fmt, ...)
							FORMAT_PRINTF(2, 3);
    /**
      * The message_v method is used to send an error message to
      * a suitable destination.  The `fmt' string is in the same
      * style a standard C vprintf function.  Derived classes are
      * required to supply this method.
      */
    virtual void message_v(const char *fmt, va_list ap) = 0;

    /**
      * The exit method is used to terminate execution.  It runs the
      * registered atext handlers, and then called the exit_inner metho
      * of the derived class.
      */
    void exit(int arg)                                             NORETURN;

    /**
      * The set_quitter clas smethod is uaed to specify a different
      * quitter to the default.  (The default uses stdio to ptint error
      * messages, and ::exit to leave th eprogram.
      */
    static void set_quitter(quit &arg);

    /**
      * The get_quitter class method is used to obtain the current
      * setting of the which quitter is to be used.
      */
    static quit &get_quitter();

    /**
      * The register_atexit methof is used to register an object to be
      * called when the program exits (via the quit::exit method).
      */
    void register_atexit(quit_atexit *arg);

protected:
    /**
      * The exit_inner method is used to terminate execution.  It
      * is called by the exit method, after it has called all the
      * regostered atexit handlers.  It could call the ::exit function,
      * or it could throw an exception, or some other action, but it
      * shall NOT return.  Derived classes are required to supply this
      * method.
      */
    virtual NORETURN void exit_inner(int arg) = 0;

private:
    /**
      * The atexit instance variabl is used to remember all the things
      * that have to be done at program exit.
      */
    quit_atexit_list atexit;

    /**
      * The quitter_p class variabel is sued to remember the location of
      * the current quitter.
      */
    static quit *quitter_p;
};

/**
  * The quitter pseudo variable is a reference to a quit implementation.
  * It should be used unless there is a good reason not to.  If you want
  * to change the quitter being used, use the quit::set_quitter method.
  */
#define quitter (quit::get_quitter())

#endif // LIB_QUIT_H
