//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the rfc822 class
//

#pragma implementation "rfc822"

#include <cassert>
#include <cctype>
#include <cstdlib>
#include <cstdio> // for snprintf
#include <ctime>

#include <boolean.h>
#include <input/file.h>
#include <input/remove_comments.h>
#include <output/file.h>
#include <quit.h>
#include <rcstring/accumulator.h>
#include <rfc822.h>


rfc822::~rfc822()
{
}


rfc822::rfc822()
{
    database.set_reaper();
}


rfc822::rfc822(const rfc822 &arg)
{
    database.set_reaper();
    database = arg.database;
}


rfc822 &
rfc822::operator=(const rfc822 &arg)
{
    if (this != &arg)
	database = arg.database;
    return *this;
}


rcstring
rfc822::sanitize_name(const rcstring &name)
{
    rcstring_accumulator acc;
    const char *cp = name.c_str();
    for (;;)
    {
	unsigned char c = *cp++;
	switch (c)
	{
	case 0:
	    while (acc.back() == '-')
		acc.pop_back();
	    if (acc.empty())
		acc.push_back("empty");
	    return acc.mkstr();

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	case '_':
	case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
	case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
	case 's': case 't': case 'u': case 'v': case 'w': case 'x':
	case 'y': case 'z':
	    acc.push_back(c);
	    break;

	case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
	case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
	case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
	case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
	case 'Y': case 'Z':
	    acc.push_back(tolower(c));
	    break;

	default:
	    if (!acc.empty() && acc.back() != '-')
		acc.push_back('-');
	    break;
	}
    }
}


void
rfc822::set(const rcstring &name, const rcstring &value)
{
    //
    // Make sure the name is RFC 822 compliant.
    //
    database.assign(sanitize_name(name), value);
}


void
rfc822::set_minimalist(const rcstring &name, const rcstring &value)
{
    rcstring saname = sanitize_name(name);
    if (value.empty())
	database.remove(saname);
    else
	database.assign(saname, value);
}


void
rfc822::set(const rcstring &name, bool value)
{
    set(name, bool_to_string(value));
}


void
rfc822::set(const rcstring &name, long value)
{
    char buf[20];
    snprintf(buf, sizeof(buf), "%ld", value);
    set(name, rcstring(buf));
}


void
rfc822::set(const rcstring &name, unsigned long value)
{
    char buf[20];
    snprintf(buf, sizeof(buf), "%lu", value);
    set(name, rcstring(buf));
}


const rcstring &
rfc822::get(const rcstring &name)
{
    //
    // Look in the database for the named item.  The names are all
    // case-insensitive and are stored in lower case.
    //
    rcstring lc_name(sanitize_name(name));
    rcstring *value_p = database.query(lc_name);
    if (!value_p)
    {
	//
	// Default it to the empty string.
	//
	database.assign(lc_name, rcstring());
	value_p = database.query(lc_name);
	assert(value_p);
    }

    //
    // Return the value found.
    //
    return *value_p;
}


rcstring
rfc822::get(const rcstring &name)
    const
{
    //
    // Look in the database for the named item.  The names are all
    // case-insensitive and are stored in lower case.
    //
    rcstring lc_name(sanitize_name(name));
    rcstring *value_p = database.query(lc_name);
    return (value_p ? *value_p : "");
}


long
rfc822::get_long(const rcstring &name)
{
    const rcstring &value = get(name);
    return strtol(value.c_str(), 0, 10);
}


unsigned long
rfc822::get_ulong(const rcstring &name)
{
    const rcstring &value = get(name);
    return strtoul(value.c_str(), 0, 10);
}


void
rfc822::load(input &source)
{
    //
    // We are looking for RFC822 style header lines.
    // When we see a blank line, we stop.
    //
    rcstring_accumulator name;
    rcstring_accumulator value;
    for (;;)
    {
	int c = source.getch();
	if (c < 0 || c == '\n')
	    break;
	name.clear();
	for (;;)
	{
	    switch ((unsigned char)c)
	    {
	    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
	    case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
	    case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
	    case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
	    case 'Y': case 'Z':
	    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	    case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
	    case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
	    case 's': case 't': case 'u': case 'v': case 'w': case 'x':
	    case 'y': case 'z':
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
	    case '_': case '-':
		name.push_back(c);
		c = source.getch();
		if (c < 0)
		    goto malformed;
		continue;

	    default:
		break;
	    }
	    break;
	}
	if (name.empty() || c != ':')
	{
	    malformed:
	    quitter.fatal_error
	    (
		"%s: malformed RFC 822 header line",
		source.name().c_str()
	    );
	    // NOTREACHED
	}

	//
	// Skip white space between the colon and the start of the
	// value.  (There is no way to have a value which consists
	// entirely of white space.)
	//
	for (;;)
	{
	    c = source.getch();
	    if  (c < 0 || c == '\n' || !isspace((unsigned char)c))
		break;
	}

	//
	// Read the value.
	//
	value.clear();
	for (;;)
	{
	    if (c < 0)
		break;
	    if (c == '\n')
	    {
		c = source.getch();
		if (c < 0)
		    break;
		if (c == '\n' || !isspace((unsigned char)c))
		{
		    source.ungetc(c);
		    break;
		}
		value.push_back('\n');
	    }
	    value.push_back(c);
	    c = source.getch();
	}

	//
	// Remove trailing white space.
	//
	while (isspace((unsigned char)value.back()))
	    value.pop_back();

	//
        // Stash the name-value pair in the database.  This means later
        // lines silently over-ride earlier lines.  We aren't expecting
        // duplicates, but some uses of this class can take advantage of
        // this behaviour.
	//
	set(name.mkstr(), value.mkstr());
    }
}


void
rfc822::load_from_file(const rcstring &filename)
{
    input_file in_raw(filename);
    in_raw.empty_if_absent();
    input_remove_comments in(in_raw);
    load(in);
}


void
rfc822::store(output &dest)
{
    rcstring_list names;
    database.keys(names);
    names.sort();
    for (size_t j = 0; j < names.size(); ++j)
    {
	const rcstring &name = names[j];
	dest << name.capitalize() << ": ";

	//
        // We have to make sure that the value is written in an RFC822
        // compliant format.  That means we have to be careful of
        // newlines embedded in the value embedded in the value and make
        // sure they are always followed by a space or a tab.
	//
	const rcstring &value = get(name);
	const char *cp = value.c_str();
	const char *end = cp + value.size();
	while (end > cp && isspace((unsigned char)end[-1]))
	    --end;
	while (cp < end)
	{
	    unsigned char c = *cp++;
	    dest << c;
	    if (c == '\n' && *cp != ' ' && *cp != '\t')
		dest << ' ';
	}
	dest << '\n';
    }
    dest << '\n';
}


void
rfc822::store_to_file(const rcstring &filename)
{
    output_file os(filename);
    store(os);
}


rcstring
rfc822::date()
{
    time_t now;
    time(&now);
    struct tm *now_tm = localtime(&now);
    char buffer[100];
    strftime(buffer, sizeof(buffer), "%a, %e %b %Y %H:%M:%S %z", now_tm);
    return buffer;
}


bool
rfc822::is_set(const rcstring &name)
    const
{
    //
    // Look in the database for the named item.  The names are all
    // case-insensitive and are stored in lower case.
    //
    rcstring lc_name(sanitize_name(name));
    rcstring *value_p = database.query(lc_name);
    return (value_p != 0);
}


void
rfc822::clear()
{
    database.clear();
}
