//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment checkin
//

#include <cctype>
#include <cstdlib>
#include <cstdio>
#include <pwd.h>

#include <adler32.h>
#include <careful.h>
#include <input/file.h>
#include <input/gunzip.h>
#include <mime_type.h>
#include <output/file.h>
#include <output/gzip.h>
#include <quit/atexit/unlink.h>
#include <quit.h>
#include <rfc822.h>
#include <rfc822_functor/version_search.h>
#include <rfc822_functor/version_previous.h>
#include <simple_version_tool.h>


static rcstring
build_temp_file_name(const rcstring &path)
{
    rcstring dirname = path.dirname();
    rcstring basename = path.basename();
    return dirname + "/.tmp." + basename;
}


static rcstring
get_user_name()
{
#ifndef DO_NOT_USE_GETPWUID
    struct passwd *pw = getpwuid(getuid());
    if (pw)
	return pw->pw_name;
#endif
    char buffer[20];
    snprintf(buffer, sizeof(buffer), "uid:%d", getuid());
    return buffer;
}


void
simple_version_tool::checkin(const rcstring &input_file_name)
{
    rfc822 meta_data;
    checkin(input_file_name, meta_data);
}


void
simple_version_tool::checkin(const rcstring &input_file_name,
    const rfc822 &meta_data)
{
    rfc822_functor_version_previous prev;
    bool there_are_previous_versions = !list(prev);
    rcstring version;
    if (meta_data.is_set("version"))
    {
	version = meta_data.get("version");
	if (version.empty())
	{
	    quitter.fatal_error
	    (
		"version number \"%s\" not appropriate",
		version.c_str()
	    );
	}
    }
    else
    {
	if (there_are_previous_versions)
	{
            //
            // If the old version ends with digits, add one to the
            // digits and preserve the prefix.
	    //
	    rcstring old_version = prev.get_result();
	    const char *start = old_version.c_str();
	    const char *end = start + old_version.size();
	    const char *cp = end;
	    while (cp > old_version.c_str() && isdigit((unsigned char)cp[-1]))
		--cp;
	    if (cp < end)
	    {
		int n = atoi(cp);
		char buffer[20];
		snprintf(buffer, sizeof(buffer), "%d", n + 1);
		version = rcstring(start, cp - start) + rcstring(buffer);
	    }
	}
    }
    if (version.empty())
    {
	//
	// When all else fails, start from one.
	//
	version = "1";
    }

    //
    // Make sure the new version number (or whatever it is) will be unique.
    //
    rfc822_functor_version_search snoopy(version);
    list(snoopy);
    if (snoopy.get_result())
    {
	quitter.fatal_error
	(
	    "%s: version \"%s\" already exists",
	    history_file_name.c_str(),
	    version.c_str()
	);
    }

    //
    // We have to remember to unlink the temporary file if something
    // goes wrong.  (The temporary file is not in /tmp because we want
    // to do a simple rename at the end, and that has to be on the same
    // file system.)
    //
    rcstring temp_file_name = build_temp_file_name(history_file_name);
    quitter.register_atexit(new quit_atexit_unlink(temp_file_name));

    //
    // Open the input file.
    //
    // We will need some information about it to put in the header.
    //
    input_file in(input_file_name);

    //
    // Build the header to write to the file.
    //
    rfc822 header(meta_data);
    header.set("content-type", mime_type(input_file_name));
    header.set("checksum", calculate_adler32(input_file_name));
    header.set("version", version);
    header.set("content-length", in.get_length());
    header.set("content-transfer-encoding", rcstring("8bit"));

    //
    // These things are merely defaults in the case where the user does
    // not define any corresponding meta-data herself.
    //
    if (!header.is_set("date"))
	header.set("date", rfc822::date());
    if (!header.is_set("user"))
	header.set("user", get_user_name());

    //
    // We have to scope the output instance, so that the output is all
    // written and closed before we do the rename.
    //
    {
	output_file out_u(temp_file_name);
	output_gzip out(&out_u, false);

	//
        // Each version consists of an RFC 822 header, and the
        // file contents, which can be binary.  The header has a
        // Content-Length entry, so we know exactly how long the content
        // is.  This means we can skip down the file versions when they
        // are simply placed end-to-end.
        //
        // Note that if the source file is plain text, then the
        // (uncompressed) output will be plain text, too.
	//
	header.store(out);
	out << in;

	if (there_are_previous_versions)
	{
	    //
            // Add the previous history on the end of the new file.
            // These previous versions should compress *very* well,
            // because thay are usually almost the same as the new
            // version.  Thus the gzip compression tables will apply,
            // and provide reasonable compression.
	    //
	    input_file old_hist_uncom(history_file_name);
	    input_gunzip old_history_file(&old_hist_uncom, false);
	    out << old_history_file;
	}
    }

    //
    // Now that we have the new file contents, we can replace them with
    // the new file contents.  (This does the unlink of the old file, as
    // well.)
    //
    careful_rename(temp_file_name, history_file_name);
}
