//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the tree_walk class
//

#pragma implementation "tree_walk"

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <configuration.h>
#include <path_join.h>
#include <quit.h>
#include <tree_walk.h>
#include <tree_walk/functor.h>
#include <wiki_word.h>


tree_walk::~tree_walk()
{
}


tree_walk::tree_walk(const rcstring &arg) :
    root(arg),
    sort_by_name(false)
{
    if (root.empty())
	root = "/";
}


bool
tree_walk::descend(tree_walk_functor &inspector)
{
    return descend(".", inspector);
}


bool
tree_walk::descend(const rcstring &path, tree_walk_functor &inspector)
{
    rcstring user_path(path_join(root, path));
    rcstring file_system_path(path_join(config.get_localstatedir(), user_path));

    //
    // Read all the names of the files and subdirectories.
    //
    // We don't do this with the directory open, because deep recusions
    // would run out of file descriptors.  And there is a sort bonus, too.
    //
    DIR *dp = opendir(file_system_path.c_str());
    if (!dp)
    {
	quitter.fatal_error_with_errno
	(
	    "opendir \"%s\"",
	    file_system_path.c_str()
	);
    }
    rcstring_list names;
    for (;;)
    {
	struct dirent *dep = readdir(dp);
	if (!dep)
	    break;
	const char *np = dep->d_name;
	if (np[0] == '.' && (!np[1] || (np[1] == '.' && !np[2])))
	    continue;
	names.push_back(rcstring(np));
    }
    closedir(dp);

    //
    // Sort the names as this point, because it's cheaper than sorting
    // the whole results list at the end.
    //
    if (sort_by_name)
	names.sort_nocase();

    //
    // Now see if we need to recurse.
    //
    for (size_t j = 0; j < names.size(); ++j)
    {
	//
	// Everything in the admin directory is synthesized, or so
	// we would have the users believe.
	//
	const rcstring &name(names[j]);
	if (root == "/" && name == "admin")
	    continue;

	//
        // See what sort of file is loacted at the path formed by joining
        // the directory name and the file name together with a slash.
	//
	rcstring path2(path_join(file_system_path, name));
	struct stat st;
	if (stat(path2.c_str(), &st) < 0)
	    quitter.fatal_error_with_errno("stat \"%s\"", path2.c_str());
	switch (st.st_mode & S_IFMT)
	{
	case S_IFREG:
	    //
            // Only bother remembering sensable .wiki files, all the
            // rest are internal to Quickie.
	    //
	    if (name.size() > 5 && name.ends_with(".wiki"))
	    {
		rcstring wname(name.c_str(), name.size() - 5);
		if
		(
		    wname == wiki_word_sanitizer(wname)
		&&
		    !inspector(path_join(user_path, wname), st)
		)
		    return false;
	    }
	    break;

	case S_IFDIR:
	    //
            // Only bother walking directories that have sensable names,
            // otherwise GET will not be able to get them.
	    //
	    if
    	    (
		name == wiki_word_sanitizer(name)
	    &&
		!descend(path_join(path, name), inspector)
	    )
		return false;
	    break;

	default:
	    //
	    // Throw everything else away, including symlinks.
	    //
	    break;
	}
    }
    return true;
}
