//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the user_t class
//

#pragma implementation "user_t"

#include <unistd.h>

#include <boolean.h>
#include <cgi.h>
#include <configuration.h>
#include <exists_and.h>
#include <mkdir_p.h>
#include <output.h>
#include <user.h>

//
// The single user instance for the whole program.
//
user_t user;


user_t::~user_t()
{
}


user_t::user_t() :
    filename("/dev/null"),
    slurp_settings_done(false),
    dirty(false),
    anonymous(false)
{
}


rcstring
user_t::calculate_filename(const rcstring &name)
{
    return
	(
	    config.get_localstatedir()
	+
	    "/admin/user.db/"
	+
	    name.substring(0, 1) + "/"
	+
	    name.substring(0, 2) + "/"
	+
	    name
	);
}


void
user_t::set_filename(const rcstring &name)
{
    filename = calculate_filename(name);
}


void
user_t::slurp_settings()
{
    if (slurp_settings_done)
	return;
    slurp_settings_done = true;

    //
    // Get there user name cookie, and generate the pathname of the user
    // information file from it.
    //
    rcstring name = gateway.get_cookie("quickie-user-name");
    if (name.size() < 3)
    {
	anonymous_settings();
	return;
    }

    //
    // The dot means this directory tree can never be accessed as wiki
    // pages.  This is intentional so that users' details can't be
    // leaked.
    //
    set_filename(name);
    settings.load_from_file(filename);

    //
    // Check the user's key (a secret known only to user user's browser
    // and to us) against the key on record.  We need to do this so that
    // impersonating another user requires more than just knowing their
    // user name.
    //
    rcstring key = gateway.get_cookie("quickie-user-key");
    if (settings.get("key") != key)
    {
	settings.clear();
	anonymous_settings();
	return;
    }
}


bool
user_t::login(const rcstring &name, const rcstring &password)
{
    if (name.empty())
	return false;
    slurp_settings_done = true;
    settings.clear();
    set_filename(name);
    settings.load_from_file(filename);
    if (settings.get("name") != name)
    {
	barf:
	settings.clear();
	anonymous_settings();
	slurp_settings_done = true;
	return false;
    }
    rcstring pwd1(settings.get("password"));
    rcstring pwd2(crypt(password.c_str(), pwd1.c_str()));
    if (pwd1 != pwd2)
	goto barf;
    return true;
}


void
user_t::anonymous_settings()
{
    filename = "/dev/null";
    settings.set("name", rcstring("anonymous"));
    settings.set("edit", config.get_anonymous_user_edit());
    dirty = false;
    anonymous = true;
}


rcstring
user_t::get(const rcstring &name)
{
    slurp_settings();
    return settings.get(name);
}


bool
user_t::get_bool(const rcstring &name, bool dflt)
{
    bool result = string_to_bool(get(name), dflt);
    settings.set(name, result);
    return result;
}


void
user_t::set(const rcstring &name, const rcstring &value)
{
    if (name == "name")
	set_filename(value);
    rcstring old = settings.get(name);
    if (value != old)
    {
	settings.set(name, value);
	dirty = true;
    }
}


void
user_t::set(const rcstring &name, bool value)
{
    set(name, bool_to_string(value));
}


bool
user_t::may_edit()
{
    return get_bool("edit", false);
}


void
user_t::flush()
{
    if (dirty)
    {
	mkdir_above(filename);
	settings.store_to_file(filename);
	dirty = false;
    }
}


void
user_t::default_settings()
{
    settings.clear();
    slurp_settings_done = true;
    set("edit", config.get_default_user_edit());
}


bool
user_t::exists(const rcstring &name)
{
    return exists_and_is_readable(calculate_filename(name));
}

bool
user_t::is_anonymous()
{
    slurp_settings();
    return anonymous;
}


rcstring
user_t::get_name()
{
    return get("name");
}


static void
print_cookie(output &out, const rcstring &name, const rcstring &value)
{
    out << "Set-Cookie: quickie-" << name << "=" << value << "; path=/\n";
}


void
user_t::set_cookie(output &out)
{
    print_cookie(out, "user-name", get("name"));
    print_cookie(out, "user-key", get("key"));
}
