//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/user.cc
//

#ifndef LIB_USER_H
#define LIB_USER_H

#pragma interface "user_t"

#include <rfc822.h>

/**
  * The user_t class is used to represent the current state of the
  * remote user's details and permissions.
  */
class user_t
{
public:
    /**
      * The destructor.
      *
      * Note that the flish method is NOT called in the destructor.
      * This is intetional.
      */
    virtual ~user_t();

    /**
      * The default constructor.
      */
    user_t();

    /**
      * The may_edit method is sued to determine whether or not the
      * current user is allowed to edit wiki pages.
      */
    bool may_edit();

    /**
      * The flush method is used to write all of the settinngs back out
      * to file, if something has changed recently.
      *
      * @note
      *     This method is NOT called by the destructor.
      *     This is intentional.
      */
    void flush();

    /**
      * The exists class method is used to determine the the named user
      * exists (actually, the file holding the settings and attributes
      * and permission sof the user).
      *
      * @param name
      *     The name of the user to test for.
      */
    static bool exists(const rcstring &name);

    /**
      * The default_settings method is used to load the settings
      * instance variables with the configuration information (in the
      * Quickie config file) into the settings instance variable.
      *
      * The configurations fields named "default-user-*" will be used.
      * The dirty flag will be set.
      */
    void default_settings();

    /**
      * The set method is used to assign a value to the named setting.
      * If the setting was not present, or the value differs from the
      * existing value, the dirty flag will be set.
      *
      * @param name
      *     The name of the setting (property or permission).
      * @param value
      *     The value to be used.
      */
    void set(const rcstring &name, const rcstring &value);

    /**
      * The is_anonymous instance variabel is used to remember whether
      * or not the current user is anonymous.  The opposite of "logged
      * in".
      */
    bool is_anonymous();

    /**
     * The get_name method is used to get the name of the current user.
     */
    rcstring get_name();

    /**
      * The login method is used to attempt to log in the named user.
      *
      * @param name
      *     The user name given to the dialogue
      * @param password
      *     The passowrd given to the dialogue
      * @returns
      *     boolean; true if logged in successfully, false if
      *     unsuccessful.  (Do NOT tell this if it was the username or
      *     the password at fault, because that leaks information.)
      */
    bool login(const rcstring &name, const rcstring &password);

    /**
      * The set_cookie method is used to write the cookie sttings to the
      * output.
      *
      * @param os
      *     The output stream to write on.
      */
    void set_cookie(output &os);

private:
    /**
      * The filename instance variable is used to remember the actual
      * file system path of the file used to store the user's details
      * and permissions.
      */
    rcstring filename;

    /**
      * The settings instance variable is used to remember the in-memory
      * cache of the on-disk file of the user's details and permissions.
      */
    rfc822 settings;

    /**
      * The slurp_settings_done instance variable is used to remember
      * whether or not the slurp_settings method has set been called to
      * fill in the settings instance variable.
      */
    bool slurp_settings_done;

    /**
      * The dirty instance variable is used to remember whether or not
      * settings instance variabel has been changed since it was read
      * in.  This means it needs to be written out again, via the flush
      * method.
      */
    bool dirty;

    /**
      * The anonymous instance variable is used to rememeber whether
      * or not the current user is anonymous.  This is the oppisite of
      * "logged in".  Not meaningful until the slurp_settings method has
      * been called.
      */
    bool anonymous;

    /**
      * The slurp_settings method is used to read the data in the file
      * (as indicated by the filename instance variable).
      *
      * @note
      *     This method assumes you have called the set_filename method
      *     already.
      */
    void slurp_settings();

    /**
      * The anonymous_settings method is used to load the settings
      * instance variables with the configuration information (in the
      * Quickie config file) into the settings instance variable.
      *
      * The configurations fields named "anonymous-user-*" will be used.
      * the dirty flag will be cleared.
      */
    void anonymous_settings();

    /**
      * The get method is used to fetch the named property out of the
      * settings symbol table.
      *
      * @param name
      *     The name of the setting (property or permission).
      * @returns
      *     The value of the setting, or the empty string if it is not
      *     (yet) present.
      */
    rcstring get(const rcstring &name);

    /**
      * The get_bool method is used to fetch the named boolean property
      * out of the settings symbol table.
      *
      * @param name
      *     The name of the setting (property or permission).
      * @param dflt
      *     The default value to be used if the setting wasnot already
      *     present.  (If the default is used, it will be stored in the
      *     settings, and the dirty flag will be set.)
      * @returns
      *     The value of the setting.
      */
    bool get_bool(const rcstring &name, bool dflt = false);

    /**
      * The set method is used to assign a boolean value to the name
      * setting.  If the setting was not present, or the value differs
      * from the existing value, the dirty flag will be set.
      *
      * @param name
      *     The name of the setting (property or permission).
      * @param value
      *     The value to be used.
      */
    void set(const rcstring &name, bool value);

    /**
      * The set_filename method is used to remember the on-disk name
      * of the file used to hold the user's settings and properties and
      * permissions.
      *
      * @param name
      *     The name of the user.
      */
    void set_filename(const rcstring &name);

    /**
      * The calculate_filename method is used to calculate the on-disk
      * name of the file used to hold the user's settings and properties
      * and permissions.
      *
      * @param name
      *     The name of the user.
      */
    static rcstring calculate_filename(const rcstring &name);

    /**
      * The copy constructor.  Do not use.
      */
    user_t(const user_t &);

    /**
      * The assignment operator.  Do not use.
      */
    user_t &operator=(const user_t &);
};

/**
  * The user variable is the single instance for the whole system.
  */
extern user_t user;

#endif // LIB_USER_H
