//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the wiki_to class
//

#pragma implementation "wiki_to"

#include <cctype>

#include <input.h>
#include <output.h>
#include <rcstring/accumulator.h>
#include <wiki_to.h>


wiki_to::~wiki_to()
{
    isp = 0;
    osp = 0;
}


wiki_to::wiki_to(input &arg1, output &arg2) :
    isp(&arg1),
    osp(&arg2)
{
}


void
wiki_to::italics()
{
    //
    // The ''italics'' indicator is ambiguous.  To see if it is the
    // beginning or the ending, we have to look and see if there is an
    // open italics somewhere in the stack.
    //
    if (!stack.contains(wiki_mode::type_italics))
    {
	if (stack.empty())
	    paragraph_begin();
	wiki_mode *wmp = italics_factory();
	stack.push_back(wmp);
	wmp->begin();
    }
    else
    {
	for (;;)
	{
	    wiki_mode *wmp = stack.pop_back();
	    bool stop = (wmp->get_type() == wiki_mode::type_italics);
	    wmp->end();
	    delete wmp;
	    if (stop)
		break;
	}
    }
}


void
wiki_to::bold()
{
    //
    // The __bold__ indicator is ambiguous.  To see if it is the
    // beginning or the ending, we have to look and see if there is an
    // open bold somewhere in the stack.
    //
    if (!stack.contains(wiki_mode::type_bold))
    {
	if (stack.empty())
	    paragraph_begin();
	wiki_mode *wmp = bold_factory();
	stack.push_back(wmp);
	wmp->begin();
    }
    else
    {
	for (;;)
	{
	    wiki_mode *wmp = stack.pop_back();
	    if (!wmp)
		break;
	    bool stop = (wmp->get_type() == wiki_mode::type_bold);
	    wmp->end();
	    delete wmp;
	    if (stop)
		break;
	}
    }
}


void
wiki_to::typewriter_begin()
{
    if (stack.empty())
	paragraph_begin();
    wiki_mode *wmp = typewriter_factory();
    stack.push_back(wmp);
    wmp->begin();
}


void
wiki_to::typewriter_end()
{
    for (;;)
    {
	wiki_mode *wmp = stack.pop_back();
	if (!wmp)
	    break;
	bool stop = (wmp->get_type() == wiki_mode::type_typewriter);
	wmp->end();
	delete wmp;
	if (stop)
	    break;
    }
}


void
wiki_to::anchor_begin()
{
    //
    // We can only have one anchor mode at a some.  It's an HTML
    // constraint, and its also semantically much more simple.
    //
    if (stack.contains(wiki_mode::type_anchor))
	anchor_end();
    if (stack.empty())
	paragraph_begin();

    //
    // Read the text of the link, so that we know what we are dealing
    // with.  It could be a local wiki page, or it could be an external
    // link.
    //
    rcstring_accumulator text;
    for (;;)
    {
	int ic = isp->getch();
	if (ic < 0)
	{
            // If we see unterminated text, pretend that they gave us a
            // suitable termination character.
	    ic = ']';
	}
	unsigned char c = ic;
	if (c == '"' || !isprint(c))
	{
	    isp->ungetc(c);
	    c = ']';
	}
	if (c == ']')
	{
	    // the URL for the link and the text for the link are the same.
	    rcstring s = text.mkstr();
	    wiki_mode *wmp = anchor_factory(s);
	    wmp->begin();
	    puts(s);
	    wmp->end();
	    delete wmp;
	    return;
	}
	if (c == '|')
	{
	    rcstring s = text.mkstr();
	    wiki_mode *wmp = anchor_factory(s);
	    stack.push_back(wmp);
	    wmp->begin();
	    return;
	}
	text.push_back(c);
    }
}


void
wiki_to::anchor_end()
{
    for (;;)
    {
	wiki_mode *wmp = stack.pop_back();
	if (!wmp)
	    break;
	bool stop = (wmp->get_type() == wiki_mode::type_anchor);
	wmp->end();
	delete wmp;
	if (stop)
	    break;
    }
}


void
wiki_to::heading_begin()
{
    //
    // We can only have one heading mode at a some.  It's an HTML
    // constraint, and its also semantically much more simple.
    //
    stack_clear();

    //
    // Count the number of exclamation marks.
    // This tells us the heading depth.
    //
    int n = 1;
    int c = isp->getch();
    for (;;)
    {
	if (c != '!')
	    break;
        c = isp->getch();
	++n;
    }

    //
    // Skip any white space between the exclamation make and the start
    // of the heading text.
    //
    for (;;)
    {
	if (c < 0)
	    break;
	if (c == '\n' || !isspace((unsigned char)c))
	{
	    isp->ungetc(c);
	    break;
	}
        c = isp->getch();
    }

    //
    // Now push the mode onto the stack.
    //
    wiki_mode *wmp = heading_factory(n);
    stack.push_back(wmp);
    wmp->begin();
}


void
wiki_to::heading_end()
{
    for (;;)
    {
	wiki_mode *wmp = stack.pop_back();
	if (!wmp)
	    break;
	bool stop = (wmp->get_type() == wiki_mode::type_heading);
	wmp->end();
	delete wmp;
	if (stop)
	    break;
    }
}


void
wiki_to::pre()
{
    //
    // HTML considers <PRE> to be outside any paragraphs, so we
    // make sure all open modes are neatly ended.
    //
    stack_clear();

    wiki_mode *wmp = preformatted_factory();
    wmp->begin();

    int n = 0;
    for(;;)
    {
	int c = isp->getch();
	if (c == '}')
	{
	    ++n;
	    if (n >= 3)
		break;
	}
	else
	{
	    while (n > 0)
	    {
		putch('}');
		--n;
	    }
	    if (c < 0)
		break;
	    putch(c);
	}
    }
    wmp->end();
    delete wmp;
}


void
wiki_to::putch_para(char c)
{
    if (stack.empty())
	paragraph_begin();
    putch(c);
}


void
wiki_to::stack_clear()
{
    while (!stack.empty())
    {
	wiki_mode *wmp = stack.pop_back();
	wmp->end();
	delete wmp;
    }
}


void
wiki_to::paragraph_begin()
{
    stack_clear();

    wiki_mode *wmp = paragraph_factory();
    stack.push_back(wmp);
    wmp->begin();
}


void
wiki_to::paragraph_end()
{
    stack_clear();
}


void
wiki_to::translate()
{
    bool start_of_line = true;
    for (;;)
    {
	bool start_of_line_was = start_of_line;
	start_of_line = false;
	int ic = isp->getch();
	if (ic < 0)
	{
	    //
	    // Make sure all open modes are neatly ended.
	    //
	    stack_clear();
	    if (!osp->start_of_line())
		putch('\n');
	    return;
	}
	unsigned char c = ic;
	switch (c)
	{
	case '!':
	    if (!start_of_line_was)
		putch_para(c);
	    else
		heading_begin();
	    break;

	case '#':
	    if (!start_of_line_was)
	    {
		putch_para(c);
		break;
	    }
	    // This is a comment.
	    // It lasts until the end of the line.
	    for (;;)
	    {
		ic = isp->getch();
		if (ic < 0 || ic == '\n')
		    break;
	    }
	    start_of_line = true;
	    break;

	case '[':
	    //
	    // Start of a link, if only one.
	    //
	    anchor_begin();
	    break;

	case ']':
	    //
            // If there is an anchor in progress, this is its finish,
            // otherwise it's simply a right square bracket.
	    //
	    if (stack.contains(wiki_mode::type_anchor))
		anchor_end();
	    else
		putch_para(']');
	    break;

	case '\'':
	    //
	    // Two of these in a row is the start or finish of italics,
	    // otherwise its simply a single quote.
	    //
	    ic = isp->getch();
	    if (ic < 0)
	    {
		putch_para('\'');
		break;
	    }
	    c = ic;
	    if (c != '\'')
	    {
		putch_para('\'');
		isp->ungetc(c);
		break;
	    }
	    italics();
	    break;

	case '_':
	    //
	    // Two of these in a row is the start or finish of bold.
	    //
	    ic = isp->getch();
	    if (ic < 0)
	    {
		putch_para('_');
		break;
	    }
	    c = ic;
	    if (c != '_')
	    {
		putch_para('_');
		isp->ungetc(c);
		break;
	    }
	    // This is the start or finish of bold.
	    bold();
	    break;

	case '{':
	    //
	    // Two of these in a row is the start of mono-spaced text.
	    // Three in a row is <PRE> markup.
	    //
	    ic = isp->getch();
	    if (ic < 0)
	    {
		putch_para('{');
		break;
	    }
	    c = ic;
	    if (c != '{')
	    {
		putch_para('{');
		isp->ungetc(c);
		break;
	    }
	    ic = isp->getch();
	    if (ic < 0)
	    {
		putch_para('{');
		putch_para('{');
		break;
	    }
	    c = ic;
	    if (c == '{')
	    {
		pre();
		break;
	    }
	    isp->ungetc(c);
	    typewriter_begin();
	    break;

	case '}':
            //
            // If there is mono-space mode present, two of these in a
            // row end the monospace text.
	    //
	    ic = isp->getch();
	    if (ic < 0)
	    {
		putch_para('}');
		break;
	    }
	    c = ic;
	    if (c != '}')
	    {
		putch_para('}');
		isp->ungetc(c);
		break;
	    }
	    typewriter_end();
	    break;

	case '\n':
	    //
            // Newlines end any open heading, except that
            // newline-whitespace is a heading continuation line.
	    //
	    start_of_line = true;
	    if (stack.contains(wiki_mode::type_heading))
	    {
		ic = isp->getch();
		if (ic >= 0)
		    isp->ungetc(ic);
		if (ic != ' ' && ic != '\t')
		    heading_end();
	    }

	    //
	    // Two newlines in a row is a paragraph break.
	    //
	    ic = isp->getch();
	    if (ic < 0)
	    {
		if (!stack.empty())
		    putch_para('\n');
		break;
	    }
	    c = ic;
	    if (c != '\n')
	    {
		if (!stack.empty())
		    putch_para('\n');
		isp->ungetc(c);
		break;
	    }
	    // new paragraph
	    paragraph_end();
	    break;

	case '\r':
	    // Just throw these away.
	    // FIXME: need to handle this better
	    break;

	default:
	    putch_para(c);
	    break;
	}
    }
}


void
wiki_to::puts(const rcstring &s)
{
    for (size_t j = 0; j < s.size(); ++j)
	putch_para(s[j]);
}
