//
//	quickie - a small fast C++ Wiki Wiki
//	Copyright (C) 2005 Peter Miller
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/wiki_to.cc
//

#ifndef LIB_WIKI_TO_H
#define LIB_WIKI_TO_H

#pragma interface "wiki_to"

#include <wiki_mode/stack.h>

class input; // forward
class output; // forward

/**
  * The wiki_to abstract class is used to represent the parse state of a
  * translator from wiki mark-up to some other mark-up.
  *
  * Italics is represented as ''doubled single quotes''.
  * Bold is represented as __doubled underscores__.
  * Typewrite font is represented as {{doubled braces}}.
  * A paragraph break is represented as a blank line.
  */
class wiki_to
{
public:
    /**
      * The destructor.
      */
    virtual ~wiki_to();

    /**
      * The constructor.
      *
      * @param is
      *     The input file to read.
      * @param os
      *     The output stream to write.  It's probably a file, but it
      *     could be anything.
      */
    wiki_to(input &is, output &os);

    /**
      * The translate method is used to translate the input wiki mark-up
      * into the outpur HTML mark-up.
      *
      * The input file and output file should be examined on return from
      * this function to see if they have an error.  If this method sees
      * and error it will return quietly.
      *
      * Note that not <HTML/> or <BODY/> elements are emitted.  It is
      * expected that the surrounding context will do this.  The idea
      * is that this makes it easy to produce a document which also
      * includes a heading portion, a sidebar portion and a footer
      * portion.
      */
    void translate();

    /**
      * The look_for_heading_one method is used to read the first line
      * of the wiki file, if it is a heading line, and translate it into
      * HTML.  It is useful for extracting page titles, etc.
      *
      * @returns
      *     Translated HTML text, or the empty string of not heading was
      *     found at the beginning of input.
      *
      * @note
      *     You must call this method before you call the translate method,
      *     or nothing will happen.
      */
    rcstring look_for_heading_one();

protected:
    /**
      * The paragraph_factory method is used to create a new paragraph
      * mode instance.
      *
      * @returns
      *     pointer to wiki_mode instance in dynamic memory, delete it
      *     when you are done with it.
      */
    virtual wiki_mode *paragraph_factory() = 0;

    /**
      * The italics_factory method is used to create a new italics mode
      * instance.
      *
      * @returns
      *     pointer to wiki_mode instance in dynamic memory, delete it
      *     when you are done with it.
      */
    virtual wiki_mode *italics_factory() = 0;

    /**
      * The bold_factory method is used to create a new bold mode
      * instance.
      *
      * @returns
      *     pointer to wiki_mode instance in dynamic memory, delete it
      *     when you are done with it.
      */
    virtual wiki_mode *bold_factory() = 0;

    /**
      * The typewriter_factory method is used to create a new typewriter
      * mode instance.
      *
      * @returns
      *     pointer to wiki_mode instance in dynamic memory, delete it
      *     when you are done with it.
      */
    virtual wiki_mode *typewriter_factory() = 0;

    /**
      * The anchor_factory method is used to create a new anchor
      * mode instance.
      *
      * @param destination
      *     The text of the destination of the link.
      * @returns
      *     pointer to wiki_mode instance in dynamic memory, delete it
      *     when you are done with it.
      */
    virtual wiki_mode *anchor_factory(const rcstring &destination) = 0;

    /**
      * The heading_factory method is used to create a new heading
      * mode instance.
      *
      * @param level
      *     The nesting level of the heading (1 outermost, 8 innermost).
      * @returns
      *     pointer to wiki_mode instance in dynamic memory, delete it
      *     when you are done with it.
      */
    virtual wiki_mode *heading_factory(int level) = 0;

    /**
      * The preformatted_factory method is used to create a new
      * preformatted mode instance.
      *
      * @returns
      *     pointer to wiki_mode instance in dynamic memory, delete it
      *     when you are done with it.
      */
    virtual wiki_mode *preformatted_factory() = 0;

    /**
      * The self_factory method is used to create a new translator,
      * based on some of this class' own state.
      */
    virtual wiki_to *self_factory(input &is, output &os) = 0;

private:
    /**
      * The isp instance variable is used to remember the input stream
      * to read for the .wiki text to be translated.
      */
    input *isp;

protected:
    /**
      * The osp instance variable is used to remember the output stream
      * to write with the translated .html text.
      */
    output *osp;

    /**
      * The putch method is used to write a character to the output.
      * Each destination format will need its own personalised
      * encodings.
      *
      * @param c
      *     The character to output.
      */
    virtual void putch(char c) = 0;

    /**
      * The putch_para method is used to add a normal character to a
      * paragraph.  This is done by calling paragraph_begin of the mode
      * stack is empty, and then calling the virtual putch method.
      */
    void putch_para(char c);

    /**
      * The stack_clear method is used to close all open modes.
      */
    void stack_clear();

private:
    /**
      * The stack instance variable is used to track the open modes at
      * any point in the parse.  This allows reasonal behaviour (and
      * valid HTML output) in the face of incorrect wiki mark-up.
      */
    wiki_mode_stack stack;

    /**
      * The italics method is used to handle the appearance of doubled
      * single quotes.  This is ambiguous, it could mean the beginning or
      * the ending, depending on context.
      */
    void italics();

    /**
      * The bold method is used to handle the appearance of doubled
      * underscores.  This is ambiguous, it could mean the beginning or
      * the ending, depending on context.
      */
    void bold();

    /**
      * The typewriter_begin method is used to process the beginning of
      * a mono-spaced font portion of text.
      */
    void typewriter_begin();

    /**
      * The typewriter_end method is used to process the ending of a
      * mono-spaced font portion of text.
      */
    void typewriter_end();

    /**
      * The paragraph_begin method is used to process the beginning of
      * a paragraph of text.
      */
    void paragraph_begin();

    /**
      * The paragraph_end method is used to process the ending of a
      * paragraph of text.
      */
    void paragraph_end();

    /**
      * The anchor_begin method is used to process the beginning of
      * an anchor (a HTML hyper link).
      */
    void anchor_begin();

    /**
      * The anchor_end method is used to process the ending of an
      * anchor (a HTML hyper link).
      */
    void anchor_end();

    /**
      * The heading_begin method is used to process the beginning of a
      * heading.
      */
    void heading_begin();

    /**
      * The heading_end method is used to process the ending of a
      * heading.
      */
    void heading_end();

    /**
      * The pre method is used to process {{{ markup }}} into <PRE>
      * markup </PRE>.  There is no way to quote anything inside.
      * Minimal HTML insulation is provided.
      */
    void pre();

    /**
      * The puts method is used to emit a string through the putch method.
      */
    void puts(const rcstring &s);

    /**
      * The default constructor.  Do not use.
      */
    wiki_to();

    /**
      * The copy constructor.  Do not use.
      */
    wiki_to(const wiki_to &);

    /**
      * The assignment operator.  Do not use.
      */
    wiki_to &operator=(const wiki_to &);
};

#endif // LIB_WIKI_TO_H
